#!/usr/bin/env python3
# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Dialog for editing download subfolder structure and file renaming
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"

import copy
import datetime
import logging
import os
import webbrowser
from collections import OrderedDict
from collections.abc import Sequence
from enum import IntEnum, auto

from PyQt5.QtCore import QSignalMapper, QSize, Qt, pyqtSignal, pyqtSlot
from PyQt5.QtGui import (
    QBrush,
    QColor,
    QFont,
    QFontMetrics,
    QKeyEvent,
    QMouseEvent,
    QResizeEvent,
    QStandardItem,
    QSyntaxHighlighter,
    QTextCharFormat,
    QTextCursor,
    QTextDocument,
    QWheelEvent,
)
from PyQt5.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QFrame,
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QStyle,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)
from sortedcontainers import SortedList

import raphodo.generatename as gn
from raphodo.constants import (
    CustomColors,
    NameGenerationType,
    PrefPosition,
    PresetClass,
    PresetPrefType,
)
from raphodo.generatenameconfig import (
    DICT_SUBFOLDER_L0,
    FILENAME,
    JOB_CODE,
    METADATA,
    PHOTO_RENAME_MENU_DEFAULTS,
    PHOTO_RENAME_MENU_DEFAULTS_CONV,
    PHOTO_SUBFOLDER_MENU_DEFAULTS,
    PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV,
    SEPARATOR,
    SEQUENCES,
    TEXT,
    VIDEO_RENAME_MENU_DEFAULTS,
    VIDEO_RENAME_MENU_DEFAULTS_CONV,
    VIDEO_SUBFOLDER_MENU_DEFAULTS,
    VIDEO_SUBFOLDER_MENU_DEFAULTS_CONV,
    filter_subfolder_prefs,
)
from raphodo.prefs.preferences import (
    DownloadsTodayTracker,
    Preferences,
    match_pref_list,
)
from raphodo.rpdfile import FileType, Photo, SamplePhoto, SampleVideo, Video
from raphodo.ui.messagewidget import MessageWidget
from raphodo.ui.viewutils import (
    standardMessageBox,
    translateDialogBoxButtons,
    translateMessageBoxButtons,
)
from raphodo.utilities import remove_last_char_from_list_str


class PrefEditor(QTextEdit):
    """
    File renaming and subfolder generation preference editor
    """

    prefListGenerated = pyqtSignal()

    def __init__(self, subfolder: bool, parent=None) -> None:
        """
        :param subfolder: if True, the editor is for editing subfolder generation
        """

        super().__init__(parent)
        self.subfolder = subfolder

        self.user_pref_list = []  # type: list[str]
        self.user_pref_colors = []  # type: list[str]

        self.heightMin = 0
        self.heightMax = 65000
        # Start out with about 4 lines in height:
        self.setMinimumHeight(QFontMetrics(self.font()).lineSpacing() * 5)
        self.document().documentLayout().documentSizeChanged.connect(
            self.wrapHeightToContents
        )

    def wrapHeightToContents(self) -> None:
        """
        Adjust the text area size to show contents without vertical scrollbar

        Derived from:
        http://stackoverflow.com/questions/11851020/a-qwidget-like-qtextedit-that-wraps-its-height-automatically-to-its-contents/11858803#11858803
        """

        docHeight = self.document().size().height() + 5
        if (
            self.heightMin <= docHeight <= self.heightMax
            and docHeight > self.minimumHeight()
        ):
            self.setMinimumHeight(docHeight)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """
        Automatically select a pref value if it was clicked in
        :param event:  the mouse event
        """

        super().mousePressEvent(event)
        if event.button() == Qt.LeftButton:
            position = self.textCursor().position()
            pref_pos, start, end, left_start, left_end = self.locatePrefValue(position)

            if pref_pos == PrefPosition.on_left:
                start = left_start
                end = left_end
            if pref_pos != PrefPosition.not_here:
                cursor = self.textCursor()
                cursor.setPosition(start)
                cursor.setPosition(end + 1, QTextCursor.KeepAnchor)
                self.setTextCursor(cursor)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Automatically select pref values when navigating through the document.

        Suppress the return / enter key.

        :param event: the key press event
        """

        key = event.key()
        if key in (Qt.Key_Enter, Qt.Key_Return, Qt.Key_Tab):
            return

        cursor = self.textCursor()  # type: QTextCursor

        if cursor.hasSelection() and key in (Qt.Key_Left, Qt.Key_Right):
            # Pass the key press on and let the selection deselect
            pass
        elif key in (
            Qt.Key_Left,
            Qt.Key_Right,
            Qt.Key_Home,
            Qt.Key_End,
            Qt.Key_PageUp,
            Qt.Key_PageDown,
            Qt.Key_Up,
            Qt.Key_Down,
        ):
            # Navigation key was pressed

            # Was ctrl key pressed too?
            ctrl_key = event.modifiers() & Qt.ControlModifier

            selection_start = selection_end = -1

            # This event is called before the cursor is moved, so
            # move the cursor as if it would be moved
            if key == Qt.Key_Right and not cursor.atEnd():
                if ctrl_key:
                    cursor.movePosition(QTextCursor.WordRight)
                else:
                    cursor.movePosition(QTextCursor.Right)
            elif key == Qt.Key_Left and not cursor.atStart():
                if ctrl_key:
                    cursor.movePosition(QTextCursor.WordLeft)
                else:
                    cursor.movePosition(QTextCursor.Left)
            elif key == Qt.Key_Up:
                cursor.movePosition(QTextCursor.Up)
            elif key == Qt.Key_Down:
                cursor.movePosition(QTextCursor.Down)
            elif key in (Qt.Key_Home, Qt.Key_PageUp):
                if ctrl_key or key == Qt.Key_PageUp:
                    cursor.movePosition(QTextCursor.StartOfBlock)
                else:
                    cursor.movePosition(QTextCursor.StartOfLine)
            elif key in (Qt.Key_End, Qt.Key_PageDown):
                if ctrl_key or key == Qt.Key_PageDown:
                    cursor.movePosition(QTextCursor.EndOfBlock)
                else:
                    cursor.movePosition(QTextCursor.EndOfLine)

            # Get position of where the cursor would move to
            position = cursor.position()

            # Determine if there is a pref value to the left or at that position
            pref_pos, start, end, left_start, left_end = self.locatePrefValue(position)
            if pref_pos == PrefPosition.on_left:
                selection_start = left_start
                selection_end = left_end + 1
            elif pref_pos == PrefPosition.at:
                selection_start = end + 1
                selection_end = start
            elif pref_pos == PrefPosition.positioned_in:
                if key == Qt.Key_Left or key == Qt.Key_Home:
                    # because moving left, position the cursor on the left
                    selection_start = end + 1
                    selection_end = start
                else:
                    # because moving right, position the cursor on the right
                    selection_start = start
                    selection_end = end + 1

            if selection_end >= 0 and selection_start >= 0:
                cursor.setPosition(selection_start)
                cursor.setPosition(selection_end, QTextCursor.KeepAnchor)
                self.setTextCursor(cursor)
                return

        super().keyPressEvent(event)

    def locatePrefValue(self, position: int) -> tuple[PrefPosition, int, int, int, int]:
        """
        Determine where pref values are relative to the position passed.

        :param position: some position in text, e.g. cursor position
        :return: enum indicating where prefs are found and their start and end
         positions. Return positions are -1 if not found.
        """

        start = end = -1
        left_start = left_end = -1
        pref_position = PrefPosition.not_here
        b = self.highlighter.boundaries
        if not (len(b)):
            return (pref_position, start, end, left_start, left_end)

        index = b.bisect_left((position, 0))
        # Special cases
        if index == 0:
            # At or to the left of the first pref value
            if b[0][0] == position:
                pref_position = PrefPosition.at
                start, end = b[0]
        elif index == len(b):
            # To the right of or in the last pref value
            if position <= b[-1][1]:
                start, end = b[-1]
                pref_position = PrefPosition.positioned_in
            elif b[-1][1] == position - 1:
                left_start, left_end = b[-1]
                pref_position = PrefPosition.on_left
        else:
            left = b[index - 1]
            right = b[index]

            at = right[0] == position
            to_left = left[1] == position - 1
            if at and to_left:
                pref_position = PrefPosition.on_left_and_at
                start, end = right
                left_start, left_end = left
            elif at:
                pref_position = PrefPosition.at
                start, end = right
            elif to_left:
                pref_position = PrefPosition.on_left
                left_start, left_end = left
            elif position <= left[1]:
                pref_position = PrefPosition.positioned_in
                start, end = b[index - 1]

        return (pref_position, start, end, left_start, left_end)

    def displayPrefList(self, pref_list: Sequence[str]) -> None:
        p = pref_list
        values = []
        for i in range(0, len(pref_list), 3):
            try:
                value = f"<{self.pref_mapper[(p[i], p[i + 1], p[i + 2])]}>"
            except KeyError:
                if p[i] == SEPARATOR:
                    value = SEPARATOR
                else:
                    assert p[i] == TEXT
                    value = p[i + 1]
            values.append(value)

        self.document().clear()
        cursor = self.textCursor()  # type: QTextCursor
        cursor.insertText("".join(values))

    def insertPrefValue(self, pref_value: str) -> None:
        cursor = self.textCursor()  # type: QTextCursor
        cursor.insertText(f"<{pref_value}>")

    def _setHighlighter(self) -> None:
        self.highlighter = PrefHighlighter(
            list(self.string_to_pref_mapper.keys()), self.pref_color, self.document()
        )

        # when color coding of text in the editor is complete,
        # generate the preference list
        self.highlighter.blockHighlighted.connect(self.generatePrefList)

    def setPrefMapper(
        self, pref_mapper: dict[tuple[str, str, str], str], pref_color: dict[str, str]
    ) -> None:
        self.pref_mapper = pref_mapper
        self.string_to_pref_mapper = {value: key for key, value in pref_mapper.items()}

        self.pref_color = pref_color
        self._setHighlighter()

    def _parseTextFragment(self, text_fragment) -> None:
        if self.subfolder:
            text_fragments = text_fragment.split(os.sep)
            for index, text_fragment in enumerate(text_fragments):
                if text_fragment:
                    self.user_pref_list.extend([TEXT, text_fragment, ""])
                    self.user_pref_colors.append("")
                if index < len(text_fragments) - 1:
                    self.user_pref_list.extend([SEPARATOR, "", ""])
                    self.user_pref_colors.append("")
        else:
            self.user_pref_list.extend([TEXT, text_fragment, ""])
            self.user_pref_colors.append("")

    def _addColor(self, pref_defn: str) -> None:
        self.user_pref_colors.append(self.pref_color[pref_defn])

    @pyqtSlot()
    def generatePrefList(self) -> None:
        """
        After syntax highlighting has completed, use its findings
        to generate the user's pref list
        """

        text = self.document().toPlainText()
        b = self.highlighter.boundaries

        self.user_pref_list = pl = []  # type: list[str]
        self.user_pref_colors = []  # type: list[str]

        # Handle any text at the very beginning
        if b and b[0][0] > 0:
            text_fragment = text[: b[0][0]]
            self._parseTextFragment(text_fragment)

        if len(b) > 1:
            for index, item in enumerate(b[1:]):
                start, end = b[index]
                # Add + 1 to start to remove the opening <
                pl.extend(self.string_to_pref_mapper[text[start + 1 : end]])
                # Add + 1 to start to include the closing >
                self._addColor(text[start : end + 1])

                text_fragment = text[b[index][1] + 1 : item[0]]
                self._parseTextFragment(text_fragment)

        # Handle the final pref value
        if b:
            start, end = b[-1]
            # Add + 1 to start to remove the opening <
            pl.extend(self.string_to_pref_mapper[text[start + 1 : end]])
            # Add + 1 to start to include the closing >
            self._addColor(text[start : end + 1])
            final = end + 1
        else:
            final = 0

        # Handle any remaining text at the very end (or the complete string if there are
        # no pref definition values)
        if final < len(text):
            text_fragment = text[final:]
            self._parseTextFragment(text_fragment)

        assert len(self.user_pref_colors) == len(self.user_pref_list) / 3
        self.prefListGenerated.emit()


class PrefHighlighter(QSyntaxHighlighter):
    """
    Highlight non-text preference values in the editor
    """

    blockHighlighted = pyqtSignal()

    def __init__(
        self,
        pref_defn_strings: list[str],
        pref_color: dict[str, str],
        document: QTextDocument,
    ) -> None:
        super().__init__(document)

        # Where detected preference values start and end:
        # [(start, end), (start, end), ...]
        self.boundaries = SortedList()

        pref_defns = (f"<{pref}>" for pref in pref_defn_strings)
        self.highlightingRules = []
        for pref in pref_defns:
            format = QTextCharFormat()
            format.setForeground(QBrush(QColor(pref_color[pref])))
            self.highlightingRules.append((pref, format))

    def find_all(self, text: str, pref_defn: str):
        """
        Find all occurrences of a preference definition in the text
        :param text: text to search
        :param pref_defn: the preference definition
        :return: yield the position in the document's text
        """
        if not len(pref_defn):
            return  # do not use raise StopIteration as it is Python 3.7 incompatible
        start = 0
        while True:
            start = text.find(pref_defn, start)
            if start == -1:
                # do not use raise StopIteration as it is Python 3.7 incompatible
                return
            yield start
            start += len(pref_defn)

    def highlightBlock(self, text: str) -> None:
        # Recreate the preference value from scratch
        self.boundaries = SortedList()

        for expression, format in self.highlightingRules:
            for index in self.find_all(text, expression):
                length = len(expression)
                self.setFormat(index, length, format)
                self.boundaries.add((index, index + length - 1))

        self.blockHighlighted.emit()


def make_subfolder_menu_entry(prefs: tuple[str]) -> str:
    """
    Create the text for a menu / combobox item

    :param prefs: single pref item, with title and elements
    :return: item text
    """

    desc = prefs[0]
    elements = prefs[1:]
    # Translators: this text appears in menus and combo boxes. It displays the
    # description of an item, and its elements.
    # Translators: %(variable)s represents Python code, not a plural of the term
    # variable. You must keep the %(variable)s untranslated, or the program will
    # crash.
    return _("%(description)s - %(elements)s") % dict(
        description=desc, elements=os.sep.join(elements)
    )


def make_rename_menu_entry(prefs: tuple[str]) -> str:
    """
    Create the text for a menu / combobox item

    :param prefs: single pref item, with title and elements
    :return: item text
    """

    desc = prefs[0]
    elements = prefs[1]
    # Translators: this text appears in menus and combo boxes. It displays the
    # description of an item, and its elements.
    # Translators: %(variable)s represents Python code, not a plural of the term
    # variable. You must keep the %(variable)s untranslated, or the program will
    # crash.
    return _("%(description)s - %(elements)s") % dict(
        description=desc, elements=elements
    )


class PresetComboBoxNegativeOffset(IntEnum):
    REMOVE_ALL_CUSTOM = auto()
    REMOVE_CUSTOM = auto()
    SAVE_NEW_CUSTOM = auto()


class PresetComboBox(QComboBox):
    """
    Combox box displaying built-in presets, custom presets,
    and some commands relating to preset management.

    Used in dialog window used to edit name generation and
    also in the rename files panel.
    """

    def __init__(
        self,
        prefs: Preferences,
        preset_names: list[str],
        preset_type: PresetPrefType,
        edit_mode: bool,
        parent=None,
    ) -> None:
        """
        :param prefs: program preferences
        :param preset_names: list of custom preset names
        :param preset_type: one of photo rename, video rename,
         photo subfolder, or video subfolder
        :param edit_mode: if True, the combo box is being displayed
         in an edit dialog window, else it's being displayed in the
         file rename panel
        :param parent: parent widget
        """

        super().__init__(parent)
        self.edit_mode = edit_mode
        self.prefs = prefs

        self.preset_edited = False
        self.new_preset = False

        self.preset_type = preset_type

        if preset_type == PresetPrefType.preset_photo_subfolder:
            self.builtin_presets = PHOTO_SUBFOLDER_MENU_DEFAULTS
        elif preset_type == PresetPrefType.preset_video_subfolder:
            self.builtin_presets = VIDEO_SUBFOLDER_MENU_DEFAULTS
        elif preset_type == PresetPrefType.preset_photo_rename:
            self.builtin_presets = PHOTO_RENAME_MENU_DEFAULTS
        else:
            assert preset_type == PresetPrefType.preset_video_rename
            self.builtin_presets = VIDEO_RENAME_MENU_DEFAULTS

        self._setup_entries(preset_names)

    def _setup_entries(self, preset_names: list[str]) -> None:
        idx = 0

        if self.edit_mode:
            for pref in self.builtin_presets:
                self.addItem(make_subfolder_menu_entry(pref), PresetClass.builtin)
                idx += 1
        else:
            for pref in self.builtin_presets:
                self.addItem(pref[0], PresetClass.builtin)
                idx += 1

        if not len(preset_names):
            # preset_separator bool is used to indicate the existence of
            # a separator in the combo box that is used to distinguish
            # custom from built-in prests
            self.preset_separator = False
        else:
            self.preset_separator = True

            self.insertSeparator(idx)
            idx += 1

            for name in preset_names:
                self.addItem(name, PresetClass.custom)
                idx += 1

        self.insertSeparator(idx)

        if self.edit_mode:
            self.addItem(_("Save New Custom Preset..."), PresetClass.new_preset)
            self.remove_custom_preset_generic_title = _("Remove Custom Preset...")
            self.remove_custom_preset_title = _(
                'Remove Custom Preset "%(preset_name)s"...'
            )
            self.addItem(
                self.remove_custom_preset_generic_title, PresetClass.remove_preset
            )
            self.addItem(_("Remove All Custom Presets..."), PresetClass.remove_all)
            self.setRemoveCustomEnabled(enabled=False)
            self.setRemoveAllCustomEnabled(bool(len(preset_names)))
        else:
            self.addItem(_("Custom..."), PresetClass.start_editor)

    def resetEntries(self, preset_names: list[str]) -> None:
        assert not self.edit_mode
        self.clear()
        self._setup_entries(preset_names)

    def addCustomPreset(self, text: str) -> None:
        """
        Adds a new custom preset name to the comboxbox and sets the
        combobox to display it.

        :param text: the custom preset name
        """

        assert self.edit_mode
        if self.new_preset or self.preset_edited:
            self.resetPresetList()
        if not self.preset_separator:
            self.insertSeparator(len(self.builtin_presets))
            self.preset_separator = True
        idx = len(self.builtin_presets) + 1
        self.insertItem(idx, text, PresetClass.custom)
        self.setCurrentIndex(idx)

    def removeAllCustomPresets(self, no_presets: int) -> None:
        assert self.edit_mode
        assert self.preset_separator
        start = len(self.builtin_presets)
        if self.new_preset:
            start += 2
        elif self.preset_edited:
            self.resetPresetList()
        end = start + no_presets
        for row in range(end, start - 1, -1):
            self.removeItem(row)
        self.preset_separator = False

    def removeCustomPreset(self, name: str) -> None:
        assert self.edit_mode
        assert self.preset_separator
        row = self.findText(name)
        self.removeItem(row)

    def setPresetNew(self) -> None:
        assert self.edit_mode
        assert not self.preset_edited
        if self.new_preset:
            return
        item_text = _("(New Custom Preset)")
        self.new_preset = True
        self.insertItem(0, item_text, PresetClass.edited)
        self.insertSeparator(1)
        self.setCurrentIndex(0)

    def setPresetEdited(self, text: str) -> None:
        """
        Adds a new entry at the top of the combobox indicating that the current
        preset has been edited.

        :param text: the preset name to use
        """

        assert self.edit_mode
        assert not self.new_preset
        assert not self.preset_edited
        item_text = _("%s (edited)") % text
        self.insertItem(0, item_text, PresetClass.edited)
        self.insertSeparator(1)
        self.addItem(_('Update Custom Preset "%s"') % text, PresetClass.update_preset)
        self.preset_edited = True
        self.setCurrentIndex(0)

    def resetPresetList(self) -> None:
        """
        Removes the combo box first line 'Preset name (edited)' or
        '(New Custom Preset)', and its separator
        """

        assert self.edit_mode
        assert self.new_preset or self.preset_edited
        # remove combo box first line 'Preset name (edited)' or '(New Custom Preset)'
        self.removeItem(0)
        # remove separator
        self.removeItem(0)
        # remove Update Preset
        if self.preset_edited:
            index = self.count() - 1
            self.removeItem(index)
        self.preset_edited = self.new_preset = False

    def _getRowItem(self, offset: int) -> QStandardItem:
        assert self.edit_mode
        # Model is a QStandardItemModel
        model = self.model()
        count = self.count()
        row = count - offset - 1 if self.preset_edited else count - offset
        return model.item(row, 0)

    def _setRowEnabled(self, enabled: bool, offset: int) -> None:
        item = self._getRowItem(offset=offset)
        if not enabled:
            item.setFlags(Qt.NoItemFlags)
        else:
            item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)

    def setRemoveAllCustomEnabled(self, enabled: bool) -> None:
        self._setRowEnabled(
            enabled=enabled, offset=PresetComboBoxNegativeOffset.REMOVE_ALL_CUSTOM
        )

    def setRemoveCustomEnabled(self, enabled: bool) -> None:
        self._setRowEnabled(
            enabled=enabled, offset=PresetComboBoxNegativeOffset.REMOVE_CUSTOM
        )

    def updateRemoveCustomName(self, name: str) -> None:
        item = self._getRowItem(offset=PresetComboBoxNegativeOffset.REMOVE_CUSTOM)
        if name:
            text = self.remove_custom_preset_title % dict(preset_name=name)
        else:
            text = self.remove_custom_preset_generic_title
        item.setText(text)

    def setSaveNewCustomPresetEnabled(self, enabled: bool) -> None:
        self._setRowEnabled(
            enabled=enabled, offset=PresetComboBoxNegativeOffset.SAVE_NEW_CUSTOM
        )

    def getComboBoxIndex(self, preset_index: int) -> int:
        """
        Calculate the index into the combo box list allowing for the separator
        and other elements in the list of entries the user sees

        :param preset_index: the preset index (built-in & custom)
        :return: the index into the actual combobox entries including
         any separators etc.
        """

        if self.edit_mode and (self.new_preset or self.preset_edited):
            preset_index += 2
        if preset_index < len(self.builtin_presets):
            return preset_index
        else:
            assert self.preset_separator
            return preset_index + 1

    def getPresetIndex(self, combobox_index: int) -> int:
        """
        Opposite of getComboBoxIndex: calculates the preset index based on the
        given combo box index (which includes separators etc.)
        :param combobox_index: the index into the combobox entries the user sees
        :return: the index into the presets (built-in & custom)
        """

        if self.edit_mode and (self.new_preset or self.preset_edited):
            combobox_index -= 2
        if combobox_index < len(self.builtin_presets):
            return combobox_index
        else:
            assert self.preset_separator
            return combobox_index - 1


class CreatePreset(QDialog):
    """
    Very simple dialog window that allows user entry of new preset name.

    Save button is disabled when the current name entered equals an existing
    preset name or is empty.
    """

    def __init__(self, existing_custom_names: list[str], parent=None) -> None:
        super().__init__(parent)

        self.existing_custom_names = existing_custom_names

        self.setModal(True)

        title = _("Save New Custom Preset - Rapid Photo Downloader")
        self.setWindowTitle(title)

        self.name = QLineEdit()
        metrics = QFontMetrics(QFont())
        self.name.setMinimumWidth(metrics.width(title))
        self.name.textEdited.connect(self.nameEdited)
        flayout = QFormLayout()
        flayout.addRow(_("Preset Name:"), self.name)

        buttonBox = QDialogButtonBox()
        buttonBox.addButton(QDialogButtonBox.Cancel)  # type: QPushButton
        self.saveButton = buttonBox.addButton(QDialogButtonBox.Save)  # type: QPushButton
        self.saveButton.setEnabled(False)
        translateDialogBoxButtons(buttonBox)
        buttonBox.rejected.connect(self.reject)
        buttonBox.accepted.connect(self.accept)

        layout = QVBoxLayout()
        layout.addLayout(flayout)
        layout.addWidget(buttonBox)

        self.setLayout(layout)

    @pyqtSlot(str)
    def nameEdited(self, name: str):
        enabled = False
        if len(name) > 0:
            enabled = name not in self.existing_custom_names
        self.saveButton.setEnabled(enabled)

    def presetName(self) -> str:
        """
        :return: the name of the name the user wants to save the preset as
        """

        return self.name.text()


def make_sample_rpd_file(
    sample_job_code: str,
    prefs: Preferences,
    generation_type: NameGenerationType,
    sample_rpd_file: Photo | Video | None = None,
) -> Photo | Video:
    """
    Create a sample_rpd_file used for displaying to the user an example of their
    file renaming preference in action on a sample file.

    :param sample_job_code: sample of a Job Code
    :param prefs: user preferences
    :param generation_type: one of photo/video filenames/subfolders
    :param sample_rpd_file: sample RPDFile that will possibly be overwritten
     with new values
    :return: sample RPDFile
    """

    downloads_today_tracker = DownloadsTodayTracker(
        day_start=prefs.day_start, downloads_today=prefs.downloads_today
    )
    sequences = gn.Sequences(downloads_today_tracker, prefs.stored_sequence_no)
    if sample_rpd_file is not None:
        if sample_rpd_file.metadata is None:
            logging.error(
                "Sample file %s is missing its metadata", sample_rpd_file.full_file_name
            )
            sample_rpd_file = None
        else:
            sample_rpd_file.sequences = sequences
            sample_rpd_file.download_start_time = datetime.datetime.now()

    if sample_rpd_file is None:
        if generation_type in (
            NameGenerationType.photo_name,
            NameGenerationType.photo_subfolder,
        ):
            sample_rpd_file = SamplePhoto(sequences=sequences)
        else:
            sample_rpd_file = SampleVideo(sequences=sequences)

    sample_rpd_file.job_code = sample_job_code
    sample_rpd_file.strip_characters = prefs.strip_characters
    if sample_rpd_file.file_type == FileType.photo:
        sample_rpd_file.generate_extension_case = prefs.photo_extension
    else:
        sample_rpd_file.generate_extension_case = prefs.video_extension

    return sample_rpd_file


class EditorCombobox(QComboBox):
    """
    Regular combobox, but ignores the mouse wheel
    """

    def wheelEvent(self, event: QWheelEvent) -> None:
        event.ignore()


class PrefDialog(QDialog):
    """
    Dialog window to allow editing of file renaming and subfolder generation
    """

    def __init__(
        self,
        pref_defn: OrderedDict,
        user_pref_list: list[str],
        generation_type: NameGenerationType,
        prefs: Preferences,
        sample_rpd_file: Photo | Video | None = None,
        max_entries=0,
        parent=None,
    ) -> None:
        """
        Set up dialog to display all its controls based on the preference
        definition being used.

        :param pref_defn: definition of possible preference choices, i.e.
         one of DICT_VIDEO_SUBFOLDER_L0, DICT_SUBFOLDER_L0, DICT_VIDEO_RENAME_L0
         or DICT_IMAGE_RENAME_L0
        :param user_pref_list: the user's actual rename / subfolder generation
         preferences
        :param generation_type: enum specifying what kind of name is being edited
         (one of photo filename, video filename, photo subfolder, video subfolder)
        :param prefs: program preferences
        :param exiftool_process: daemon exiftool process
        :param sample_rpd_file: a sample photo or video, whose contents will be
         modified (i.e. don't pass a live RPDFile)
        :param max_entries: maximum number of entries that will be displayed
         to the user (in a menu, for example)
        """

        super().__init__(parent)

        self.setModal(True)

        self.generation_type = generation_type

        match generation_type:
            case NameGenerationType.photo_subfolder:
                self.setWindowTitle(_("Photo Subfolder Generation Editor"))
                self.preset_type = PresetPrefType.preset_photo_subfolder
                self.builtin_pref_lists = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV
                self.builtin_pref_names = [
                    make_subfolder_menu_entry(pref)
                    for pref in PHOTO_SUBFOLDER_MENU_DEFAULTS
                ]
                self.preset_type_hr = _("photo subfolder generation")
            case NameGenerationType.video_subfolder:
                self.setWindowTitle(_("Video Subfolder Generation Editor"))
                self.preset_type = PresetPrefType.preset_video_subfolder
                self.builtin_pref_lists = VIDEO_SUBFOLDER_MENU_DEFAULTS_CONV
                self.builtin_pref_names = [
                    make_subfolder_menu_entry(pref)
                    for pref in VIDEO_SUBFOLDER_MENU_DEFAULTS
                ]
                self.preset_type_hr = _("video subfolder generation")
            case NameGenerationType.photo_name:
                self.setWindowTitle(_("Photo Renaming Editor"))
                self.preset_type = PresetPrefType.preset_photo_rename
                self.builtin_pref_lists = PHOTO_RENAME_MENU_DEFAULTS_CONV
                self.builtin_pref_names = [
                    make_rename_menu_entry(pref) for pref in PHOTO_RENAME_MENU_DEFAULTS
                ]
                self.preset_type_hr = _("photo renaming")
            case _:
                self.setWindowTitle(_("Video Renaming Editor"))
                self.preset_type = PresetPrefType.preset_video_rename
                self.builtin_pref_lists = VIDEO_RENAME_MENU_DEFAULTS_CONV
                self.builtin_pref_names = [
                    make_rename_menu_entry(pref) for pref in VIDEO_RENAME_MENU_DEFAULTS
                ]
                self.preset_type_hr = _("video renaming")

        self.prefs = prefs
        self.max_entries = max_entries

        # Cache custom preset name and pref lists
        self.updateCachedPrefLists()

        self.current_custom_name = None

        # Setup values needed for name generation

        self.sample_rpd_file = make_sample_rpd_file(
            sample_rpd_file=sample_rpd_file,
            sample_job_code=self.prefs.most_recent_job_code(missing=_("Job Code")),
            prefs=self.prefs,
            generation_type=generation_type,
        )

        # Setup widgets and helper values

        # Translators: please do not modify or leave out html formatting tags like <i>
        # and <b>. These are used to format the text the user sees.
        warning_msg = _(
            '<b><font color="red">Warning:</font></b> <i>There is insufficient data to '
            "fully generate the name. Please use other renaming options.</i>"
        )

        self.is_subfolder = generation_type in (
            NameGenerationType.photo_subfolder,
            NameGenerationType.video_subfolder,
        )

        if self.is_subfolder:
            # Translators: please do not modify, change the order of or leave out html
            # formatting tags like <i> and <b>. These are used to format the text the
            # users sees. In this case, the </i> really is supposed to come before the
            # <i>.
            # Translators: %(variable)s represents Python code, not a plural of the term
            # variable. You must keep the %(variable)s untranslated, or the program will
            # crash.
            subfolder_msg = _(
                "The character</i> %(separator)s <i>creates a new subfolder level."
            ) % dict(separator=os.sep)
            # Translators: please do not modify, change the order of or leave out html
            # formatting tags like <i> and <b>. These are used to format the text the
            # users sees In this case, the </i> really is supposed to come before the
            # <i>.
            # Translators: %(variable)s represents Python code, not a plural of the term
            # variable. You must keep the %(variable)s untranslated, or the program will
            # crash.
            subfolder_first_char_msg = _(
                "There is no need start or end with the folder separator </i> "
                "%(separator)s<i>, because it is added automatically."
            ) % dict(separator=os.sep)
            messages = (warning_msg, subfolder_msg, subfolder_first_char_msg)
        else:
            # Translators: please do not modify or leave out html formatting tags like
            # <i> and <b>. These are used to format the text the users sees
            unique_msg = _(
                '<b><font color="red">Warning:</font></b> <i>Unique filenames may not '
                "be generated. Make filenames unique by using Sequence values.</i>"
            )
            messages = (warning_msg, unique_msg)

        self.messageWidget = MessageWidget(messages=messages)

        self.editor = PrefEditor(subfolder=self.is_subfolder)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        sizePolicy.setVerticalStretch(1)
        self.editor.setSizePolicy(sizePolicy)

        self.editor.prefListGenerated.connect(
            self.updateExampleFilenameAndComboBoxIndex
        )

        # Generated subfolder / file name example
        self.example = QLabel()

        # Combobox with built-in and user defined presets
        self.preset = PresetComboBox(
            prefs=prefs,
            preset_names=self.preset_names,
            preset_type=self.preset_type,
            edit_mode=True,
        )
        self.preset.activated.connect(self.presetComboItemActivated)

        glayout = QGridLayout()
        presetLabel = QLabel(_("Preset:"))
        exampleLabel = QLabel(_("Example:"))

        glayout.addWidget(presetLabel, 0, 0)
        glayout.addWidget(self.preset, 0, 1)
        glayout.addWidget(exampleLabel, 1, 0)
        glayout.addWidget(self.example, 1, 1)
        glayout.setColumnStretch(1, 10)

        layout = QVBoxLayout()
        self.setLayout(layout)

        layout.addLayout(glayout)
        layout.addSpacing(int(QFontMetrics(QFont()).height() / 2))
        layout.addWidget(self.editor)
        layout.addWidget(self.messageWidget)

        self.area = QScrollArea()
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy.setVerticalStretch(10)
        self.area.setSizePolicy(sizePolicy)
        self.area.setFrameShape(QFrame.NoFrame)
        layout.addWidget(self.area)

        gbSizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)

        areaWidget = QWidget()
        areaLayout = QVBoxLayout()
        areaWidget.setLayout(areaLayout)
        areaWidget.setSizePolicy(gbSizePolicy)

        self.area.setWidget(areaWidget)
        self.area.setWidgetResizable(True)

        areaLayout.setContentsMargins(0, 0, 0, 0)

        self.pushButtonSizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        self.mapper = QSignalMapper(self)
        self.widget_mapper = dict()  # type: dict[str, QComboBox| QLabel]
        self.pref_mapper = dict()  # type: dict[tuple[str, str, str], str]
        self.pref_color = dict()  # type: dict[str, str]

        titles = [title for title in pref_defn if title not in (TEXT, SEPARATOR)]
        pref_colors = {title: color.value for title, color in zip(titles, CustomColors)}
        self.filename_pref_color = pref_colors[FILENAME]

        for title in titles:
            title_i18n = _(title)
            color = pref_colors[title]
            level1 = pref_defn[title]
            gb = QGroupBox(title_i18n)
            gb.setSizePolicy(gbSizePolicy)
            gb.setFlat(True)
            areaLayout.addWidget(gb)
            gLayout = QGridLayout()
            gb.setLayout(gLayout)
            if level1 is None:
                assert title == JOB_CODE
                widget1 = QLabel(" " + title_i18n)
                widget2 = self.makeInsertButton()
                self.widget_mapper[title] = widget1
                self.mapper.setMapping(widget2, title)
                self.pref_mapper[(title, "", "")] = title_i18n
                self.pref_color[f"<{title_i18n}>"] = color
                gLayout.addWidget(self.makeColorCodeLabel(color), 0, 0)
                gLayout.addWidget(widget1, 0, 1)
                gLayout.addWidget(widget2, 0, 2)
            elif title == METADATA:
                elements = []
                data = []
                for element in level1:
                    element_i18n = _(element)
                    level2 = level1[element]
                    if level2 is None:
                        elements.append(element_i18n)
                        data.append([METADATA, element, ""])
                        self.pref_mapper[(METADATA, element, "")] = element_i18n
                        self.pref_color[f"<{element_i18n}>"] = color
                    else:
                        for e in level2:
                            e_i18n = _(e)
                            # Translators: appears in a combobox, e.g. Image Date (YYYY)
                            item = _("{choice} ({variant})").format(
                                choice=element_i18n, variant=e_i18n
                            )
                            elements.append(item)
                            data.append([METADATA, element, e])
                            self.pref_mapper[(METADATA, element, e)] = item
                            self.pref_color[f"<{item}>"] = color
                widget1 = EditorCombobox()
                for element, data_item in zip(elements, data):
                    widget1.addItem(element, data_item)
                widget2 = self.makeInsertButton()
                widget1.currentTextChanged.connect(self.mapper.map)
                self.mapper.setMapping(widget2, title)
                self.mapper.setMapping(widget1, title)
                self.widget_mapper[title] = widget1
                gLayout.addWidget(self.makeColorCodeLabel(color), 0, 0)
                gLayout.addWidget(widget1, 0, 1)
                gLayout.addWidget(widget2, 0, 2)
            else:
                for row, level1 in enumerate(pref_defn[title]):
                    widget1 = EditorCombobox()
                    level1_i18n = _(level1)
                    items = (
                        _("{choice} ({variant})").format(
                            choice=level1_i18n, variant=_(element)
                        )
                        for element in pref_defn[title][level1]
                    )
                    data = (
                        [title, level1, element] for element in pref_defn[title][level1]
                    )
                    for item, data_item in zip(items, data):
                        widget1.addItem(item, data_item)
                        self.pref_mapper[tuple(data_item)] = item
                        self.pref_color[f"<{item}>"] = color
                    widget2 = self.makeInsertButton()
                    widget1.currentTextChanged.connect(self.mapper.map)

                    self.mapper.setMapping(widget2, level1)
                    self.mapper.setMapping(widget1, level1)
                    self.widget_mapper[level1] = widget1
                    gLayout.addWidget(self.makeColorCodeLabel(color), row, 0)
                    gLayout.addWidget(widget1, row, 1)
                    gLayout.addWidget(widget2, row, 2)

        self.mapper.mapped[str].connect(self.choiceMade)

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Cancel | QDialogButtonBox.Ok | QDialogButtonBox.Help
        )
        self.helpButton = buttonBox.button(QDialogButtonBox.Help)  # type: QPushButton
        self.helpButton.clicked.connect(self.helpButtonClicked)
        self.helpButton.setToolTip(_("Get help online..."))
        translateDialogBoxButtons(buttonBox)

        buttonBox.rejected.connect(self.reject)
        buttonBox.accepted.connect(self.accept)

        layout.addWidget(buttonBox)

        self.editor.setPrefMapper(self.pref_mapper, self.pref_color)
        self.editor.displayPrefList(user_pref_list)

        self.show()
        self.setWidgetSizes()

    def helpButtonClicked(self) -> None:
        if self.generation_type in (
            NameGenerationType.photo_name,
            NameGenerationType.video_name,
        ):
            location = "#rename"
        else:
            location = "#subfoldergeneration"
        webbrowser.open_new_tab(
            f"https://damonlynch.net/rapid/documentation/{location}"
        )

    def makeInsertButton(self) -> QPushButton:
        w = QPushButton(_("Insert"))
        w.clicked.connect(self.mapper.map)
        w.setSizePolicy(self.pushButtonSizePolicy)
        return w

    def setWidgetSizes(self) -> None:
        """
        Resize widgets for enhanced visual layout
        """

        # Set the widths of the comboboxes and labels to the width of the
        # longest control
        width = max(widget.width() for widget in self.widget_mapper.values())
        for widget in self.widget_mapper.values():
            widget.setMinimumWidth(width)

        # Set the scroll area to be big enough to eliminate the horizontal scrollbar
        scrollbar_width = self.style().pixelMetric(QStyle.PM_ScrollBarExtent)
        self.area.setMinimumWidth(self.area.widget().width() + scrollbar_width)

    @pyqtSlot(str)
    def choiceMade(self, widget: str) -> None:
        """
        User has pushed one of the "Insert" buttons or selected a new value in one
        of the combo boxes.

        :param widget: widget's name, which uniquely identifies it
        """

        if widget == JOB_CODE:
            pref_value = _(JOB_CODE)
        else:
            combobox = self.widget_mapper[widget]  # type: QComboBox
            pref_value = combobox.currentText()

        self.editor.insertPrefValue(pref_value)

        # Set focus not on the control that was just used, but the editor
        self.editor.setFocus(Qt.OtherFocusReason)

    def makeColorCodeLabel(self, color: str) -> QLabel:
        """
        Generate a colored square to show beside the combo boxes / label
        :param color: color to use, e.g. #7a9c38
        :return: the square in form of a label
        """

        colorLabel = QLabel(" ")
        colorLabel.setStyleSheet("QLabel {background-color: %s;}" % color)
        size = QFontMetrics(QFont()).height()
        colorLabel.setFixedSize(QSize(size, size))
        return colorLabel

    def updateExampleFilenameAndComboBoxIndex(self) -> None:
        user_pref_list = self.editor.user_pref_list
        self.user_pref_colors = self.editor.user_pref_colors

        if not self.is_subfolder:
            self.user_pref_colors.append(self.filename_pref_color)

        self.messageWidget.setCurrentIndex(0)

        if self.is_subfolder:
            if user_pref_list:
                try:
                    user_pref_list.index(SEPARATOR)
                except ValueError:
                    # Inform the user that a subfolder separator (os.sep) is used to
                    # create subfolder levels
                    self.messageWidget.setCurrentIndex(2)
                else:
                    if (
                        user_pref_list[0] == SEPARATOR
                        or user_pref_list[-3] == SEPARATOR
                    ):
                        # inform the user that there is no need to start or finish
                        # with a subfolder separator (os.sep)
                        self.messageWidget.setCurrentIndex(3)
            else:
                # Inform the user that a subfolder separator (os.sep) is used to create
                # subfolder levels
                self.messageWidget.setCurrentIndex(2)

            changed, user_pref_list, self.user_pref_colors = filter_subfolder_prefs(
                user_pref_list, self.user_pref_colors
            )
        else:
            try:
                user_pref_list.index(SEQUENCES)
            except ValueError:
                # Inform the user that sequences can be used to make filenames unique
                self.messageWidget.setCurrentIndex(2)

        if self.generation_type == NameGenerationType.photo_name:
            self.name_generator = gn.PhotoName(user_pref_list)
        elif self.generation_type == NameGenerationType.video_name:
            self.name_generator = gn.VideoName(user_pref_list)
        elif self.generation_type == NameGenerationType.photo_subfolder:
            self.name_generator = gn.PhotoSubfolder(user_pref_list)
        else:
            assert self.generation_type == NameGenerationType.video_subfolder
            self.name_generator = gn.VideoSubfolder(user_pref_list)

        self.name_parts = self.name_generator.generate_name(
            self.sample_rpd_file, parts=True
        )
        self.showExample()
        self.updateComboBoxCurrentIndex()

    def updateComboBoxCurrentIndex(self) -> None:
        """
        Sets the combo value to match the current preference value
        """

        combobox_index, pref_list_index = self.getPresetMatch()
        if pref_list_index >= 0:
            # the editor contains an existing preset
            self.preset.setCurrentIndex(combobox_index)
            if self.preset.preset_edited or self.preset.new_preset:
                self.preset.resetPresetList()
                self.preset.setSaveNewCustomPresetEnabled(enabled=False)
            if pref_list_index >= len(self.builtin_pref_names):
                self.current_custom_name = self.preset.currentText()
                self._setRemoveCustomPresetMenuState(enabled=True)
            else:
                self.current_custom_name = None
                self._setRemoveCustomPresetMenuState(enabled=False)
        elif not (self.preset.new_preset or self.preset.preset_edited):
            if self.current_custom_name is None:
                self.preset.setPresetNew()
            else:
                self.preset.setPresetEdited(self.current_custom_name)
            self.preset.setSaveNewCustomPresetEnabled(enabled=True)
            self._setRemoveCustomPresetMenuState(enabled=False)
        else:
            self.preset.setCurrentIndex(0)
            self._setRemoveCustomPresetMenuState(enabled=False)

    def _setRemoveCustomPresetMenuState(self, enabled: bool) -> None:
        if enabled:
            self.preset.updateRemoveCustomName(name=self.current_custom_name)
            self.preset.setRemoveCustomEnabled(enabled=True)
        else:
            self.preset.updateRemoveCustomName(name="")
            self.preset.setRemoveCustomEnabled(enabled=False)

    def showExample(self) -> None:
        """
        Insert text into example widget, eliding it if necessary
        """

        user_pref_colors = self.user_pref_colors

        parts = copy.copy(self.name_parts)
        metrics = QFontMetrics(self.example.font())
        width = self.example.width() - metrics.width("…")

        # Cannot elide rich text using Qt code. Thus, elide the plain text.
        plain_text_name = "".join(parts)

        if self.is_subfolder:
            plain_text_name = self.name_generator.filter_subfolder_characters(
                plain_text_name
            )
        elided_text = metrics.elidedText(plain_text_name, Qt.ElideRight, width)
        elided = False

        while plain_text_name != elided_text:
            elided = True
            parts = remove_last_char_from_list_str(parts)
            plain_text_name = "".join(parts)
            if self.is_subfolder:
                plain_text_name = self.name_generator.filter_subfolder_characters(
                    plain_text_name
                )
            elided_text = metrics.elidedText(plain_text_name, Qt.ElideRight, width)

        colored_parts = [
            f'<span style="color: {color};">{part}</span>' if color else part
            for part, color in zip(parts, user_pref_colors)
        ]

        name = "".join(colored_parts)
        if elided:
            name = f"{name}&hellip;"

        if self.is_subfolder:
            name = self.name_generator.filter_subfolder_characters(name)

        if self.sample_rpd_file.name_generation_problem:
            self.messageWidget.setCurrentIndex(1)

        self.example.setTextFormat(Qt.RichText)
        self.example.setText(name)

    def resizeEvent(self, event: QResizeEvent) -> None:
        if self.example.text():
            self.showExample()
        super().resizeEvent(event)

    def getPrefList(self) -> list[str]:
        """
        :return: the pref list the user has specified
        """

        return self.editor.user_pref_list

    @pyqtSlot(int)
    def presetComboItemActivated(self, index: int) -> None:
        """
        Respond to user activating the Preset combo box.

        :param index: index of the item activated
        """

        preset_class = self.preset.currentData()
        if preset_class == PresetClass.new_preset:
            createPreset = CreatePreset(existing_custom_names=self.preset_names)
            if createPreset.exec():
                # User has created a new preset
                preset_name = createPreset.presetName()
                assert preset_name not in self.preset_names
                self.current_custom_name = preset_name
                self.preset.addCustomPreset(preset_name)
                self.saveNewPreset(preset_name=preset_name)
                if len(self.preset_names) == 1:
                    self.preset.setRemoveAllCustomEnabled(True)
                self.preset.setSaveNewCustomPresetEnabled(enabled=False)
            self.updateComboBoxCurrentIndex()

        elif preset_class in (PresetClass.builtin, PresetClass.custom):
            index = self.combined_pref_names.index(self.preset.currentText())
            pref_list = self.combined_pref_lists[index]
            self.editor.displayPrefList(pref_list=pref_list)
            if index >= len(self.builtin_pref_names):
                assert preset_class == PresetClass.custom
                self.movePresetToFront(index=len(self.builtin_pref_names) - index)
                self._setRemoveCustomPresetMenuState(enabled=True)
            else:
                assert preset_class == PresetClass.builtin
                self._setRemoveCustomPresetMenuState(enabled=False)

        elif preset_class == PresetClass.remove_preset:
            message = _(
                "<b>Remove Custom Preset</b><br><br>Are you sure you want to "
                'remove the custom preset "%(preset_name)s"?'
            ) % dict(preset_name=self.current_custom_name)
            msgbox = standardMessageBox(
                message=message,
                rich_text=True,
                standardButtons=QMessageBox.Yes | QMessageBox.No,
            )
            if msgbox.exec() == QMessageBox.Yes:
                if len(self.preset_names) > 1:
                    self.preset.removeCustomPreset(name=self.current_custom_name)
                    self.removeCustomPreset()
                else:
                    self._removeAllCustomPresets()
            self.updateComboBoxCurrentIndex()

        elif preset_class == PresetClass.remove_all:
            message = _(
                "<b>Remove All Custom Presets</b><br><br>Are you sure you want to "
                "remove all %(preset_type)s custom presets?"
            ) % dict(preset_type=self.preset_type_hr)
            msgbox = standardMessageBox(
                message=message,
                rich_text=True,
                standardButtons=QMessageBox.Yes | QMessageBox.No,
            )
            if msgbox.exec() == QMessageBox.Yes:
                self._removeAllCustomPresets()
            self.updateComboBoxCurrentIndex()
        elif preset_class == PresetClass.update_preset:
            self.updateExistingPreset()
            self.updateComboBoxCurrentIndex()

    def _removeAllCustomPresets(self) -> None:
        self.preset.removeAllCustomPresets(no_presets=len(self.preset_names))
        self.clearCustomPresets()
        self.preset.setRemoveAllCustomEnabled(False)

    def updateExistingPreset(self) -> None:
        """
        Updates (saves) an existing preset (assumed to be self.current_custom_name)
        with the new user_pref_list found in the editor.

        Assumes cached self.preset_names and self.preset_pref_lists represent
        current save preferences. Will update these and overwrite the relevant
        preset preference.
        """

        preset_name = self.current_custom_name
        user_pref_list = self.editor.user_pref_list
        index = self.preset_names.index(preset_name)
        self.preset_pref_lists[index] = user_pref_list
        if index > 0:
            self.movePresetToFront(index=index)
        else:
            self._updateCombinedPrefs()
            self.prefs.set_custom_presets(
                preset_type=self.preset_type,
                preset_names=self.preset_names,
                preset_pref_lists=self.preset_pref_lists,
            )

    def movePresetToFront(self, index: int) -> None:
        """
        Extracts the preset from the current list of presets and moves it
        to the front if not already there.

        Assumes cached self.preset_names and self.preset_pref_lists represent
        current save preferences. Will update these and overwrite the relevant
        preset preference.

        :param index: index into self.preset_pref_lists / self.preset_names of
         the item to move
        """

        if index == 0:
            return
        preset_name = self.preset_names.pop(index)
        pref_list = self.preset_pref_lists.pop(index)
        self.preset_names.insert(0, preset_name)
        self.preset_pref_lists.insert(0, pref_list)
        self._updateCombinedPrefs()
        self.prefs.set_custom_presets(
            preset_type=self.preset_type,
            preset_names=self.preset_names,
            preset_pref_lists=self.preset_pref_lists,
        )

    def saveNewPreset(self, preset_name: str) -> None:
        """
        Saves the current user_pref_list (retrieved from the editor) and
        saves it in the program preferences.

        Assumes cached self.preset_names and self.preset_pref_lists represent
        current save preferences. Will update these and overwrite the relevant
        preset preference.

        :param preset_name: name for the new preset
        """

        user_pref_list = self.editor.user_pref_list
        self.preset_names.insert(0, preset_name)
        self.preset_pref_lists.insert(0, user_pref_list)
        self._updateCombinedPrefs()
        self.prefs.set_custom_presets(
            preset_type=self.preset_type,
            preset_names=self.preset_names,
            preset_pref_lists=self.preset_pref_lists,
        )

    def removeCustomPreset(self) -> None:
        user_pref_list = self.editor.user_pref_list
        index = self.preset_pref_lists.index(user_pref_list)
        self.preset_names.pop(index)
        self.preset_pref_lists.pop(index)
        self.current_custom_name = None
        self._updateCombinedPrefs()
        self.prefs.set_custom_presets(
            preset_type=self.preset_type,
            preset_names=self.preset_names,
            preset_pref_lists=self.preset_pref_lists,
        )

    def clearCustomPresets(self) -> None:
        """
        Deletes all the custom presets.

        Assumes cached self.preset_names and self.preset_pref_lists represent
        current save preferences. Will update these and overwrite the relevant
        preset preference.
        """
        self.preset_names = []
        self.preset_pref_lists = []
        self.current_custom_name = None
        self._updateCombinedPrefs()
        self.prefs.set_custom_presets(
            preset_type=self.preset_type,
            preset_names=self.preset_names,
            preset_pref_lists=self.preset_pref_lists,
        )

    def updateCachedPrefLists(self) -> None:
        self.preset_names, self.preset_pref_lists = self.prefs.get_custom_presets(
            preset_type=self.preset_type
        )
        self._updateCombinedPrefs()

    def _updateCombinedPrefs(self):
        self.combined_pref_names = self.builtin_pref_names + self.preset_names
        self.combined_pref_lists = self.builtin_pref_lists + tuple(
            self.preset_pref_lists
        )

    def getPresetMatch(self) -> tuple[int, int]:
        """
        :return: tuple of the Preset combobox index and the combined pref/name list
        index, if the current user pref list matches an entry in it. Else tuple of
        (-1, -1).
        """

        index = match_pref_list(
            pref_lists=self.combined_pref_lists,
            user_pref_list=self.editor.user_pref_list,
        )
        if index >= 0:
            combobox_name = self.combined_pref_names[index]
            return self.preset.findText(combobox_name), index
        return -1, -1

    @pyqtSlot()
    def accept(self) -> None:
        """
        Slot called when the okay button is clicked.

        If there are unsaved changes, query the user if they want their changes
        saved as a new preset or if the existing preset should be updated
        """

        if self.preset.preset_edited or self.preset.new_preset:
            title = _("Save Preset - Rapid Photo Downloader")
            if self.preset.new_preset:
                message = _(
                    "<b>Do you want to save the changes in a new custom preset?</b>"
                    "<br><br>"
                    "Creating a custom preset is not required, but can help you keep "
                    "organized.<br><br>"
                    "The changes to the preferences will still be applied regardless "
                    "of whether you create a new custom preset or not."
                )
                msgBox = standardMessageBox(
                    standardButtons=QMessageBox.Yes | QMessageBox.No,
                    title=title,
                    rich_text=True,
                    message=message,
                )
                updateButton = newButton = None
            else:
                assert self.preset.preset_edited
                msgBox = QMessageBox()
                msgBox.setTextFormat(Qt.RichText)
                msgBox.setIcon(QMessageBox.Question)
                msgBox.setWindowTitle(title)
                message = _(
                    "<b>Do you want to save the changes in a custom preset?</b><br><br>"
                    "If you like, you can create a new custom preset or update the "
                    "existing custom preset.<br><br>"
                    "The changes to the preferences will still be applied regardless "
                    "of whether you save a custom preset or not."
                )
                msgBox.setText(message)
                msgBox.addButton(QMessageBox.No)
                translateMessageBoxButtons(msgBox)
                updateButton = msgBox.addButton(
                    _('Update Custom Preset "%s"') % self.current_custom_name,
                    QMessageBox.YesRole,
                )
                newButton = msgBox.addButton(
                    _("Save New Custom Preset"), QMessageBox.YesRole
                )

            choice = msgBox.exec()
            save_new = update = False
            if self.preset.new_preset:
                save_new = choice == QMessageBox.Yes
            else:
                if msgBox.clickedButton() == updateButton:
                    update = True
                elif msgBox.clickedButton() == newButton:
                    save_new = True

            if save_new:
                createPreset = CreatePreset(existing_custom_names=self.preset_names)
                if createPreset.exec():
                    # User has created a new preset
                    preset_name = createPreset.presetName()
                    assert preset_name not in self.preset_names
                    self.saveNewPreset(preset_name=preset_name)
            elif update:
                self.updateExistingPreset()

        # Check to make sure that in menus (which have a limited number of menu items)
        # that our chosen entry is displayed
        if self.max_entries:
            combobox_index, pref_list_index = self.getPresetMatch()
            if pref_list_index >= self.max_entries:
                self.updateExistingPreset()

        # Regardless of any user actions, close the dialog box
        super().accept()


if __name__ == "__main__":
    try:
        from icecream import install

        install()

    except ImportError:  # Graceful fallback if IceCream isn't installed.
        ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa
        builtins = __import__("builtins")
        setattr(builtins, "ic", ic)

    # Application development test code:

    app = QApplication([])

    app.setOrganizationName("Rapid Photo Downloader")
    app.setOrganizationDomain("damonlynch.net")
    app.setApplicationName("Rapid Photo Downloader")

    prefs = Preferences()

    # prefDialog = PrefDialog(DICT_IMAGE_RENAME_L0, PHOTO_RENAME_MENU_DEFAULTS_CONV[1],
    #                         NameGenerationType.photo_name, prefs)
    # prefDialog = PrefDialog(DICT_VIDEO_RENAME_L0, VIDEO_RENAME_MENU_DEFAULTS_CONV[1],
    #                         NameGenerationType.video_name, prefs)
    prefDialog = PrefDialog(
        DICT_SUBFOLDER_L0,
        PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV[2],
        NameGenerationType.photo_subfolder,
        prefs,
        max_entries=10,
    )
    # prefDialog = PrefDialog(
    #     DICT_VIDEO_SUBFOLDER_L0, VIDEO_SUBFOLDER_MENU_DEFAULTS_CONV[2],
    #     NameGenerationType.video_subfolder, prefs, max_entries=10
    # )
    prefDialog.show()
    app.exec_()



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Specify logging setup.

Log all messages to file log
Log messages at user specified level to console
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"

import gzip
import logging
import os
import pickle
from logging.handlers import QueueHandler, RotatingFileHandler

try:
    import colorlog

    use_colorlog = True
except ImportError:
    use_colorlog = False

from raphodo.constants import logfile_name
from raphodo.storage.storage import get_program_logging_directory

logging_format = "%(levelname)s: %(message)s"
colored_logging_format = "%(log_color)s%(levelname)-8s%(reset)s %(message)s"
log_colors = {
    "DEBUG": "cyan",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "red,bg_white",
}

logging_date_format = "%Y-%m-%d %H:%M:%S"
file_logging_format = "%(asctime)s %(levelname)s %(filename)s %(lineno)d: %(message)s"


class ZeroMQSocketHandler(QueueHandler):
    def enqueue(self, record):
        data = pickle.dumps(record.__dict__)
        self.queue.send(data)


class RotatingGzipFileHandler(RotatingFileHandler):
    def rotation_filename(self, name):
        return name + ".gz"

    def rotate(self, source, dest):
        with open(source, "rb") as sf, gzip.open(dest, "wb") as df:
            df.writelines(sf)
        os.remove(source)


def full_log_file_path():
    log_file_path = get_program_logging_directory(create_if_not_exist=True)
    if log_file_path is not None:
        log_file = os.path.join(log_file_path, logfile_name)
    else:
        # Problem: for some reason cannot create log file in standard location,
        # so create it in the home directory
        log_file = os.path.join(os.path.expanduser("~"), logfile_name)
    return log_file


def setup_main_process_logging(logging_level: int) -> logging.Logger:
    """
    Setup logging at the module level

    :param log_file_path: path where log file should be stored
    :param logging_level: logging module's logging level for console output
    :return: default logging object
    """

    log_file = full_log_file_path()
    logger = logging.getLogger()
    max_bytes = 1024 * 1024  # 1 MB
    filehandler = RotatingGzipFileHandler(log_file, maxBytes=max_bytes, backupCount=10)
    filehandler.setLevel(logging.DEBUG)
    filehandler.setFormatter(
        logging.Formatter(file_logging_format, logging_date_format)
    )
    logger.addHandler(filehandler)
    logger.setLevel(logging.DEBUG)

    consolehandler = logging.StreamHandler()
    consolehandler.set_name("console")
    if not use_colorlog:
        consolehandler.setFormatter(logging.Formatter(logging_format))
    else:
        consolehandler.setFormatter(
            colorlog.ColoredFormatter(fmt=colored_logging_format, log_colors=log_colors)
        )
    consolehandler.setLevel(logging_level)
    logger.addHandler(consolehandler)
    return logger



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Collects attributes about varieties of video formats, including how much of the file
has to be read in order to extract metadata information or generate a thumbnail.
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"


import datetime  # noqa: F401
import os
from tempfile import NamedTemporaryFile, TemporaryDirectory

import raphodo.metadata.exiftool as exiftool
from raphodo.constants import FileType
from raphodo.metadata.analysis.photoattributes import ExifToolMixin, vmtouch_output
from raphodo.metadata.metadatavideo import MetaData
from raphodo.thumbnailextractor import get_video_frame
from raphodo.utilities import datetime_roughly_equal, format_size_for_user


class VideoAttributes(ExifToolMixin):
    def __init__(
        self, full_file_name: str, ext: str, et_process: exiftool.ExifTool
    ) -> None:
        all_metadata_tags = (
            "date_time timestamp file_number width height length "
            "frames_per_second codec fourcc rotation"
        )
        super().__init__(
            FileType.video,
            full_file_name,
            et_process,
            video_metadata_scan_range,
            all_metadata_tags,
            MetaData,
        )
        self.datetime = None  # type: datetime.datetime | None
        self.file_name = full_file_name
        self.ext = ext
        self.et_process = et_process
        self.minimum_read_size_in_bytes_datetime = None  # type: int | None
        self.minimum_read_size_in_bytes_thumbnail = None  # type: int | None
        self.minimum_metadata_read_size_in_bytes_all = None  # type: int | None
        self.thumbnail_offset = 0.0

        self.assign_video_attributes(et_process)

        # Before doing further processing, understand what has already
        # been cached after simply reading the datetime metadata
        self.bytes_cached, self.total, self.in_memory = vmtouch_output(full_file_name)

        self.thumbnail = get_video_frame(full_file_name, self.thumbnail_offset)

        if self.datetime is not None:
            self.minimum_extract_for_tag(self.datetime_extract)

        if self.thumbnail:
            self.minimum_extract_for_thumbnail()

        self.minimum_extract_for_all()

    def assign_video_attributes(self, et_process: exiftool.ExifTool) -> None:
        m = MetaData(self.file_name, et_process)
        self.datetime = m.date_time(missing=None)

    def datetime_extract(self, metadata: MetaData, size_in_bytes):
        if metadata.date_time() == self.datetime:
            self.minimum_read_size_in_bytes_datetime = min(
                size_in_bytes, self.file_size
            )
            return True
        return False

    def minimum_extract_for_thumbnail(self):
        name = os.path.split(self.file_name)[1]
        with TemporaryDirectory(dir="/tmp") as tmpdirname:  # noqa: SIM117
            with open(self.file_name, "rb") as video:
                tempname = os.path.join(tmpdirname, name)
                for size_in_bytes in thumbnail_scan_range(self.file_size):
                    video.seek(0)
                    video_extract = video.read(size_in_bytes)
                    with open(tempname, "wb") as f:
                        f.write(video_extract)
                    try:
                        if (
                            get_video_frame(tempname, self.thumbnail_offset)
                            == self.thumbnail
                        ):
                            self.minimum_read_size_in_bytes_thumbnail = min(
                                size_in_bytes, self.file_size
                            )
                            break
                    except AssertionError:
                        pass

    def minimum_extract_for_tag(self, check_extract):
        with open(self.file_name, "rb") as video:
            for size_in_bytes in video_metadata_scan_range(self.file_size):
                video.seek(0)
                video_extract = video.read(size_in_bytes)
                with NamedTemporaryFile("w+b", delete=False) as f:
                    f.write(video_extract)
                    name = f.name
                metadata = MetaData(name, self.et_process)
                if check_extract(metadata, size_in_bytes):
                    os.remove(name)
                    break
                os.remove(name)

    def minimum_extract_for_all_tags(self):
        funcs = (
            "date_time timestamp file_number width height length frames_per_second "
            "codec fourcc rotation".split()
        )

        metadata = MetaData(self.file_name, self.et_process)
        for f in funcs:
            v = getattr(metadata, f)()
            if v:
                self.all_metadata_values[f] = v

        found = set()

        with open(self.file_name, "rb") as video:
            for size_in_bytes in video_metadata_scan_range(self.file_size):
                video.seek(0)
                video_extract = video.read(size_in_bytes)
                with NamedTemporaryFile("w+b", delete=False) as f:
                    f.write(video_extract)
                    name = f.name
                metadata_extract = MetaData(name, self.et_process)
                for tag in self.all_metadata_values:
                    if (
                        tag not in found
                        and getattr(metadata_extract, tag)()
                        == self.all_metadata_values[tag]
                    ):
                        found.add(tag)
                        if len(found) == len(self.all_metadata_values):
                            self.minimum_metadata_read_size_in_bytes_all = size_in_bytes
                            os.remove(name)
                            return
                os.remove(name)

    def __repr__(self):
        s = os.path.split(self.file_name)[1] if self.file_name else self.ext
        if self.datetime:
            s += f" {self.datetime}"
        if self.minimum_read_size_in_bytes_datetime:
            s += f" {self.minimum_read_size_in_bytes_datetime} (datetime)"
        if self.minimum_read_size_in_bytes_thumbnail:
            s += f" {self.minimum_read_size_in_bytes_thumbnail} (thumb)"
        if self.minimum_metadata_read_size_in_bytes_all:
            s += f" {self.minimum_metadata_read_size_in_bytes_all} (variety)"
        return s

    def __str__(self):
        if self.file_name is not None:
            s = f"{os.path.split(self.file_name)[1]}\n"
        else:
            s = self.ext
        if self.datetime:  # type: datetime.datetime
            s += "Datetime in metadata: {}\n".format(self.datetime.strftime("%c"))
            if not datetime_roughly_equal(self.datetime, self.fs_datetime):
                s += "Differs from datetime on file system: {}\n".format(
                    self.fs_datetime.strftime("%c")
                )
        else:
            s += "Datetime on file system: {}\n".format(self.fs_datetime.strftime("%c"))

        s += f"Disk cache after metadata read:\n[{self.in_memory}]\n"
        if self.minimum_read_size_in_bytes_datetime is not None:
            s += "Minimum read size to extract datetime: {} of {}\n".format(
                format_size_for_user(self.minimum_read_size_in_bytes_datetime),
                format_size_for_user(self.file_size),
            )
        if self.minimum_read_size_in_bytes_thumbnail:
            s += "Minimum read size to extract thumbnail: {} of {}\n".format(
                format_size_for_user(self.minimum_read_size_in_bytes_thumbnail),
                format_size_for_user(self.file_size),
            )
        if self.minimum_metadata_read_size_in_bytes_all is not None:
            s += "Minimum read size to extract variety of tags: {}\n".format(
                format_size_for_user(self.minimum_metadata_read_size_in_bytes_all)
            )
        else:
            s += "Could not extract variety of tags with minimal read\n"
        return s


def video_metadata_scan_range(size: int) -> iter:
    stop = 20
    for iterations, step in (
        (108, 1),
        (97, 4),
        (16, 32),
        (16, 256),
        (16, 512),
        (8, 1024),
        (8, 2048 * 4),
        (32, 2048 * 16),
        (128, 2048 * 32),
    ):
        start = stop
        stop = start + step * iterations
        yield from range(start, stop, step)
    yield size


def thumbnail_scan_range(size: int) -> iter:
    stop = 100 * 1024
    for iterations, step in (
        (10, 100 * 1024),
        (64, 1024 * 1024),
    ):
        start = stop
        stop = start + step * iterations
        yield from range(start, stop, step)
    yield size



# Copyright (C) 2015-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Class to handle exceptions related to camera access via gPhoto2 and libimobiledevice
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2015-2024, Damon Lynch."


import gphoto2 as gp

from raphodo.constants import CameraErrorCode


class CameraError(Exception):
    def __init__(self, code: CameraErrorCode) -> None:
        self.code = code

    def __repr__(self) -> str:
        return self.code.name

    def __str__(self) -> str:
        if self.code == CameraErrorCode.inaccessible:
            return "The camera is inaccessible"
        elif self.code == CameraErrorCode.locked:
            return "The camera is locked"


class CameraProblemEx(CameraError):
    """Handle gPhoto2 errors"""

    def __init__(
        self,
        code: CameraErrorCode,
        gp_exception: gp.GPhoto2Error|None = None,
        py_exception: Exception|None = None,
    ) -> None:
        super().__init__(code)
        if gp_exception is not None:
            self.gp_code = gp_exception.code
        else:
            self.gp_code = None
        self.py_exception = py_exception

    def __repr__(self) -> str:
        if self.code == CameraErrorCode.read:
            return "read error"
        elif self.code == CameraErrorCode.write:
            return "write error"
        else:
            return repr(super())

    def __str__(self) -> str:
        if self.code == CameraErrorCode.read:
            return "Could not read file from camera"
        elif self.code == CameraErrorCode.write:
            return "Could not write file from camera"
        else:
            return str(super())


class iOSDeviceError(CameraError):
    """Handle iOS Device errors"""

    def __init__(
        self,
        code: CameraErrorCode,
        imobile_error: int,
        imobile_error_output: str,
        udid: str,
        display_name: str,
    ) -> None:
        super().__init__(code)
        self.imobile_error = imobile_error
        self.imobile_error_output = imobile_error_output
        self.udid = udid
        self.display_name = display_name

    def __str__(self) -> str:

        if self.code in (
            CameraErrorCode.pair,
            CameraErrorCode.mount,
            CameraErrorCode.devicename,
        ):
            message = self.imobile_error_output.replace(
                self.udid, f"'{self.display_name}'"
            )
            if message.startswith("ERROR: "):
                message = message[7:]
            return message
        else:
            return str(super())



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Display download destination details
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"

import logging
import math
import os
from collections import defaultdict

from PyQt5.QtCore import QPoint, QRect, QSize, QStorageInfo, Qt, pyqtSlot
from PyQt5.QtGui import (
    QColor,
    QIcon,
    QMouseEvent,
    QPaintEvent,
    QPalette,
    QPixmap,  # noqa: F401
)
from PyQt5.QtWidgets import (
    QAction,
    QActionGroup,
    QApplication,
    QMenu,
    QSizePolicy,
    QSplitter,
    QStyle,
    QStylePainter,
    QWidget,
)

from raphodo.constants import (
    COLOR_RED_HTML,
    CustomColors,
    DestinationDisplayMousePos,
    DestinationDisplayTooltipState,
    DestinationDisplayType,
    DisplayingFilesOfType,
    FileType,
    NameGenerationType,
    PresetPrefType,
)
from raphodo.devices import DownloadingTo
from raphodo.generatenameconfig import (
    CUSTOM_SUBFOLDER_MENU_ENTRY_POSITION,
    DICT_SUBFOLDER_L0,
    DICT_VIDEO_SUBFOLDER_L0,
    MAX_DOWNLOAD_SUBFOLDER_MENU_ENTRIES,
    MAX_DOWNLOAD_SUBFOLDER_MENU_PRESETS,
    NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS,
    NUM_DOWNLOAD_SUBFOLDER_MENU_CUSTOM_PRESETS,
    PHOTO_SUBFOLDER_MENU_DEFAULTS,
    PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV,
    VIDEO_SUBFOLDER_MENU_DEFAULTS,
    VIDEO_SUBFOLDER_MENU_DEFAULTS_CONV,
    CustomPresetSubfolderLists,
    CustomPresetSubfolderNames,
)
from raphodo.rpdfile import FileTypeCounter, Photo, Video  # noqa: F401
from raphodo.storage.storage import StorageSpace, get_mount_size, get_path_display_name
from raphodo.ui.devicedisplay import BodyDetails, DeviceDisplay, icon_size
from raphodo.ui.nameeditor import PrefDialog, make_subfolder_menu_entry
from raphodo.ui.viewutils import darkModePixmap, paletteMidPen
from raphodo.utilities import format_size_for_user, thousands


def make_body_details(
    bytes_total: int,
    bytes_free: int,
    files_to_display: DisplayingFilesOfType,
    marked: FileTypeCounter,
    photos_size_to_download: int,
    videos_size_to_download: int,
) -> BodyDetails:
    """
    Gather the details to render for destination storage usage
    for photo and video downloads, and their backups.

    :param bytes_total:
    :param bytes_free:
    :param files_to_display:
    :param marked:
    :param photos_size_to_download:
    :param videos_size_to_download:
    :return:
    """

    bytes_total_text = format_size_for_user(bytes_total, no_decimals=0)
    existing_bytes = bytes_total - bytes_free
    existing_size = format_size_for_user(existing_bytes)

    photos = videos = photos_size = videos_size = ""

    if files_to_display != DisplayingFilesOfType.videos:
        # Translators: no_photos refers to the number of photos
        # Translators: %(variable)s represents Python code, not a plural of the term
        # variable. You must keep the %(variable)s untranslated, or the program will
        # crash.
        photos = _("%(no_photos)s Photos") % {
            "no_photos": thousands(marked[FileType.photo])
        }
        photos_size = format_size_for_user(photos_size_to_download)
    if files_to_display != DisplayingFilesOfType.photos:
        # Translators: no_videos refers to the number of videos
        # Translators: %(variable)s represents Python code, not a plural of the term
        # variable. You must keep the %(variable)s untranslated, or the program will
        # crash.
        videos = _("%(no_videos)s Videos") % {
            "no_videos": thousands(marked[FileType.video])
        }
        videos_size = format_size_for_user(videos_size_to_download)

    size_to_download = photos_size_to_download + videos_size_to_download
    comp1_file_size_sum = photos_size_to_download
    comp2_file_size_sum = videos_size_to_download
    comp3_file_size_sum = existing_bytes
    comp1_text = photos
    comp2_text = videos
    comp3_text = _("Used")
    comp4_text = _("Excess")
    comp1_size_text = photos_size
    comp2_size_text = videos_size
    comp3_size_text = existing_size

    bytes_to_use = size_to_download + existing_bytes
    percent_used = ""

    if bytes_total == 0:
        bytes_free_of_total = _("Device size unknown")
        comp4_file_size_sum = 0
        comp4_size_text = 0
        comp3_size_text = 0
    elif bytes_to_use > bytes_total:
        bytes_total_ = bytes_total
        bytes_total = bytes_to_use
        excess_bytes = bytes_to_use - bytes_total_
        comp4_file_size_sum = excess_bytes
        comp4_size_text = format_size_for_user(excess_bytes)
        # Translators: %(variable)s represents Python code, not a plural of the term
        # variable. You must keep the %(variable)s untranslated, or the program will
        # crash.
        bytes_free_of_total = _("No space free on %(size_total)s device") % dict(
            size_total=bytes_total_text
        )
    else:
        comp4_file_size_sum = 0
        comp4_size_text = 0
        bytes_free = bytes_total - bytes_to_use
        # Translators: %(variable)s represents Python code, not a plural of the term
        # variable. You must keep the %(variable)s untranslated, or the program will
        # crash.
        bytes_free_of_total = _("%(size_free)s free of %(size_total)s") % dict(
            size_free=format_size_for_user(bytes_free, no_decimals=1),
            size_total=bytes_total_text,
        )

    return BodyDetails(
        bytes_total_text=bytes_total_text,
        bytes_total=bytes_total,
        percent_used_text=percent_used,
        bytes_free_of_total=bytes_free_of_total,
        comp1_file_size_sum=comp1_file_size_sum,
        comp2_file_size_sum=comp2_file_size_sum,
        comp3_file_size_sum=comp3_file_size_sum,
        comp4_file_size_sum=comp4_file_size_sum,
        comp1_text=comp1_text,
        comp2_text=comp2_text,
        comp3_text=comp3_text,
        comp4_text=comp4_text,
        comp1_size_text=comp1_size_text,
        comp2_size_text=comp2_size_text,
        comp3_size_text=comp3_size_text,
        comp4_size_text=comp4_size_text,
        color1=QColor(CustomColors.color1.value),
        color2=QColor(CustomColors.color2.value),
        color3=QColor(CustomColors.color3.value),
        displaying_files_of_type=files_to_display,
    )


def adjusted_download_size(
    photos_size_to_download: int,
    videos_size_to_download: int,
    os_stat_device: int,
    downloading_to,
) -> tuple[int, int]:
    """
    Adjust download size to account for situations where
    photos and videos are being backed up to the same
    partition (device) they're downloaded to.

    :return: photos_size_to_download, videos_size_to_download
    """
    if os_stat_device in downloading_to:
        file_types = downloading_to[os_stat_device]
        if FileType.photo in file_types:
            photos_size_to_download = photos_size_to_download * 2
        if FileType.video in file_types:
            videos_size_to_download = videos_size_to_download * 2
    return photos_size_to_download, videos_size_to_download


class DestinationDisplay(QWidget):
    """
    Custom widget handling the display of download destinations, not including the file
    system browsing component.

    Serves a dual purpose, depending on whether photos and videos are being downloaded
    to the same file system or not:

    1. Display how much storage space the checked files will use in addition
       to the space used by existing files.

    2. Display the download destination (path), and a local menu to control subfolder
       generation.

    Where photos and videos are being downloaded to the same file system, the storage
    space display is combined into one widget, which appears in its own panel above the
    photo and video destination panels.

    Where photos and videos are being downloaded to different file systems, the combined
    display (above) is invisible, and photo and video panels have the own section in
    which to display their storage space display
    """

    photos = _("Photos")
    videos = _("Videos")
    projected_space_msg = _("Projected storage use after download")

    def __init__(
        self,
        menu: bool = False,
        file_type: FileType = None,
        parent: QWidget = None,
        rapidApp=None,
    ) -> None:
        """
        :param menu: whether to render a drop-down menu
        :param file_type: whether for photos or videos. Relevant only for menu display.
        """

        super().__init__(parent)
        self.rapidApp = rapidApp
        if rapidApp is not None:
            self.prefs = self.rapidApp.prefs
        else:
            self.prefs = None

        self.storage_space = None  # type: StorageSpace|None

        self.menu_actions = []  # type: list[QAction]
        if menu:
            pixmap = darkModePixmap(
                path=":/icons/settings.svg",
                size=QSize(100, 100),
                soften_regular_mode_color=True,
            )
            menuIcon = QIcon(pixmap)
            self.file_type = file_type
            self.createActionsAndMenu()
            self.mouse_pos = DestinationDisplayMousePos.normal
            self.tooltip_display_state = DestinationDisplayTooltipState.path
        else:
            menuIcon = None
            self.menu = None
            self.mouse_pos = None
            self.tooltip_display_state = None

        self.deviceDisplay = DeviceDisplay(parent=self, menuButtonIcon=menuIcon)
        self.deviceDisplay.widthChanged.connect(self.widthChanged)
        size = icon_size()
        self.pixmap = QIcon(":/icons/folder.svg").pixmap(QSize(size, size))  # type: QPixmap
        self.pixmap = darkModePixmap(pixmap=self.pixmap)
        self.display_name = ""
        self.photos_size_to_download = self.videos_size_to_download = 0
        self.files_to_display = None  # type: DisplayingFilesOfType|None
        self.marked = FileTypeCounter()
        self.display_type = None  # type: DestinationDisplayType|None
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)

        self.sample_rpd_file = None  # type: Photo| Video|None

        self.os_stat_device = 0  # type: int
        self._downloading_to = defaultdict(set)  # type: DownloadingTo

        self.midPen = paletteMidPen()
        self.frame_width = QApplication.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.container_vertical_scrollbar_visible = None

        self.valid = True
        self.invalidColor = QColor(COLOR_RED_HTML)
        # Translators: the lack of a period at the end is deliberate
        _("Unwritable destination")
        # Translators: the lack of a period at the end is deliberate
        _("Folder does not exist")
        # Translators: the lack of a period at the end is deliberate
        _("Insufficient storage space")


    @property
    def downloading_to(self) -> DownloadingTo:
        return self._downloading_to

    @downloading_to.setter
    def downloading_to(self, downloading_to: DownloadingTo) -> None:
        if downloading_to is not None:
            self._downloading_to = downloading_to
            # TODO determine if this is always needed here
            self.update()

    def createActionsAndMenu(self) -> None:
        self.setMouseTracking(True)
        self.menu = QMenu()

        if self.file_type == FileType.photo:
            defaults = PHOTO_SUBFOLDER_MENU_DEFAULTS
        else:
            defaults = VIDEO_SUBFOLDER_MENU_DEFAULTS

        self.subfolderGroup = QActionGroup(self)

        # Generate a list of actions with matching text entries, and place them in a
        # menu
        for index in range(MAX_DOWNLOAD_SUBFOLDER_MENU_ENTRIES):
            if index < NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS:
                menu_text = make_subfolder_menu_entry(defaults[index])
            elif index == CUSTOM_SUBFOLDER_MENU_ENTRY_POSITION:
                # Translators: Custom refers to the user choosing a non-default value
                # that they customize themselves
                menu_text = _("Custom...")
            else:
                menu_text = "Placeholder text"

            action = QAction(menu_text, self)
            action.setCheckable(True)
            action.triggered.connect(
                lambda checked, index=index: self.menuItemChosen(index)
            )

            self.subfolderGroup.addAction(action)
            self.menu_actions.append(action)

            if index == NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS:
                self.menu.addSeparator()

            self.menu.addAction(action)

    def getPresetIndex(
        self,
    ) -> tuple[int, CustomPresetSubfolderNames, CustomPresetSubfolderLists]:
        """
        Returns the index of the user's download subfolder generation config in the
        list of subfolder generation preferences, which is a combination of the built-in
        defaults and the user's custom presets.

        :return: index into the combined list of subfolder generation preferences, or -1
         if it doesn't exist in the list, as well as the user's custom presets and their
         names.
        """

        if self.file_type == FileType.photo:
            default_prefs_list = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV
            prefs_subfolder_list = self.prefs.photo_subfolder
            preset_type = PresetPrefType.preset_photo_subfolder
        else:
            default_prefs_list = VIDEO_SUBFOLDER_MENU_DEFAULTS_CONV
            prefs_subfolder_list = self.prefs.video_subfolder
            preset_type = PresetPrefType.preset_video_subfolder

        custom_preset_names, custom_preset_pref_lists = self.prefs.get_custom_presets(
            preset_type=preset_type
        )

        try:
            index = default_prefs_list.index(prefs_subfolder_list)
        except ValueError:
            try:
                index = custom_preset_pref_lists.index(prefs_subfolder_list)
            except ValueError:
                index = -1
            else:
                if index >= NUM_DOWNLOAD_SUBFOLDER_MENU_CUSTOM_PRESETS:
                    # A custom preset is in use, but due to the position of that custom
                    # preset in the list of presets, it will not be shown in the menu
                    # without being moved up in position.
                    # Move it to the beginning.
                    preset_name = custom_preset_names.pop(index)
                    pref_list = custom_preset_pref_lists.pop(index)
                    custom_preset_names.insert(0, preset_name)
                    custom_preset_pref_lists.insert(0, pref_list)

                    self.prefs.set_custom_presets(
                        preset_type=preset_type,
                        preset_names=custom_preset_names,
                        preset_pref_lists=custom_preset_pref_lists,
                    )
                    index = NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS
                else:
                    # Return the index into taking into account
                    # the length of the default presets.
                    index += NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS

        return index, custom_preset_names, custom_preset_pref_lists

    def setupMenuActions(self) -> None:
        index, preset_names, preset_pref_lists = self.getPresetIndex()
        assert index < MAX_DOWNLOAD_SUBFOLDER_MENU_ENTRIES

        action = self.menu_actions[index]  # type: QAction
        action.setChecked(True)

        # Set visibility of custom presets menu items to match how many we are
        # displaying

        for index in range(NUM_DOWNLOAD_SUBFOLDER_MENU_CUSTOM_PRESETS):
            action = self.menu_actions[NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS + index]
            if index < len(preset_names):
                action.setText(preset_names[index])
                action.setVisible(True)
            else:
                action.setVisible(False)

        # Save the custom preset list for access in self.menuItemChosen()
        self.preset_pref_lists = preset_pref_lists

    @pyqtSlot(int)
    def menuItemChosen(self, index: int) -> None:
        self.mouse_pos = DestinationDisplayMousePos.normal
        self.update()

        user_pref_list = None

        if index == CUSTOM_SUBFOLDER_MENU_ENTRY_POSITION:
            if self.file_type == FileType.photo:
                pref_defn = DICT_SUBFOLDER_L0
                pref_list = self.prefs.photo_subfolder
                generation_type = NameGenerationType.photo_subfolder
            else:
                pref_defn = DICT_VIDEO_SUBFOLDER_L0
                pref_list = self.prefs.video_subfolder
                generation_type = NameGenerationType.video_subfolder

            prefDialog = PrefDialog(
                pref_defn=pref_defn,
                user_pref_list=pref_list,
                generation_type=generation_type,
                prefs=self.prefs,
                sample_rpd_file=self.sample_rpd_file,
                max_entries=MAX_DOWNLOAD_SUBFOLDER_MENU_PRESETS,
            )
            if prefDialog.exec():
                user_pref_list = prefDialog.getPrefList()
                if not user_pref_list:
                    user_pref_list = None

        elif index >= NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS:
            assert index < CUSTOM_SUBFOLDER_MENU_ENTRY_POSITION
            user_pref_list = self.preset_pref_lists[
                index - NUM_DOWNLOAD_SUBFOLDER_BUILT_IN_PRESETS
            ]

        else:
            if self.file_type == FileType.photo:
                user_pref_list = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV[index]
            else:
                user_pref_list = VIDEO_SUBFOLDER_MENU_DEFAULTS_CONV[index]

        if user_pref_list is not None:
            logging.debug(
                "Updating %s subfolder generation preference value", self.file_type.name
            )
            if self.file_type == FileType.photo:
                self.prefs.photo_subfolder = user_pref_list
            else:
                self.prefs.video_subfolder = user_pref_list
            self.rapidApp.folder_preview_manager.change_subfolder_structure()

    def setDestination(self, path: str) -> None:
        """
        Set the downloaded destination path
        :param path: valid path
        """

        self.display_name, self.path = get_path_display_name(path)
        try:
            self.os_stat_device = os.stat(path).st_dev
        except FileNotFoundError:
            logging.error(
                "Cannot set download destination display: %s does not exist", path
            )
            self.os_stat_device = 0

        mount = QStorageInfo(path)
        bytes_total, bytes_free = get_mount_size(mount=mount)

        self.storage_space = StorageSpace(
            bytes_free=bytes_free, bytes_total=bytes_total, path=path
        )

    def setDownloadAttributes(
        self,
        marked: FileTypeCounter,
        photos_size: int,
        videos_size: int,
        files_to_display: DisplayingFilesOfType,
        display_type: DestinationDisplayType,
        merge: bool,
    ) -> None:
        """
        Set the attributes used to generate the visual display of the
        files marked to be downloaded

        :param marked: number and type of files marked for download
        :param photos_size: size in bytes of photos marked for download
        :param videos_size: size in bytes of videos marked for download
        :param files_to_display: whether displaying photos or videos or both
        :param display_type: whether showing only the header (folder only),
         usage only, or both
        :param merge: whether to replace or add to the current values
        """

        if not merge:
            self.marked = marked
            self.photos_size_to_download = photos_size
            self.videos_size_to_download = videos_size
        else:
            self.marked.update(marked)
            self.photos_size_to_download += photos_size
            self.videos_size_to_download += videos_size

        self.files_to_display = files_to_display

        self.display_type = display_type

        if self.display_type != DestinationDisplayType.usage_only:
            self.tool_tip = self.path
        else:
            self.tool_tip = self.projected_space_msg
        self.setToolTip(self.tool_tip)

        self.update()
        self.updateGeometry()

    def sufficientSpaceAvailable(self) -> bool:
        """
        Check to see that there is sufficient space with which to perform a download.

        :return: True or False value if sufficient space. Will always return False if
         the download destination is not yet set.
        """

        if self.storage_space is None:
            return False

        # allow for destinations that don't properly report their size
        if self.storage_space.bytes_total == 0:
            return True

        photos_size_to_download, videos_size_to_download = adjusted_download_size(
            photos_size_to_download=self.photos_size_to_download,
            videos_size_to_download=self.videos_size_to_download,
            os_stat_device=self.os_stat_device,
            downloading_to=self._downloading_to,
        )
        return (
            photos_size_to_download + videos_size_to_download
            < self.storage_space.bytes_free
        )

    @pyqtSlot(bool)
    def containerVerticalScrollBar(self, visible: bool) -> None:
        self.container_vertical_scrollbar_visible = visible

    def paintEvent(self, event: QPaintEvent) -> None:
        """
        Render the custom widget
        """

        painter = QStylePainter()
        painter.begin(self)

        x = 0
        y = 0
        width = self.width()

        rect = self.rect()  # type: QRect
        palette = QPalette()
        backgroundColor = palette.base().color()

        if (
            self.display_type == DestinationDisplayType.usage_only
            and QSplitter().lineWidth()
        ):
            pen = painter.pen()
            painter.setPen(backgroundColor)
            painter.drawLine(rect.topLeft(), rect.topRight())
            painter.setPen(self.midPen)
            painter.drawLine(rect.bottomLeft(), rect.bottomRight())
            painter.drawLine(rect.topLeft(), rect.bottomLeft())
            if (
                self.container_vertical_scrollbar_visible is None
                or not self.container_vertical_scrollbar_visible
            ):
                painter.drawLine(rect.topRight(), rect.bottomRight())
            painter.setPen(pen)

            w = QSplitter().lineWidth()
            rect.adjust(w, w, -w, -w)

        painter.fillRect(rect, backgroundColor)

        if self.storage_space is None:
            painter.end()
            return

        highlight_menu = self.mouse_pos == DestinationDisplayMousePos.menu

        if self.display_type != DestinationDisplayType.usage_only:
            # Render the folder icon, folder name, and the menu icon
            self.deviceDisplay.paint_header(
                painter=painter,
                x=x,
                y=y,
                width=width,
                display_name=self.display_name,
                icon=self.pixmap,
                highlight_menu=highlight_menu,
            )
            y = y + self.deviceDisplay.dc.device_name_height

        if self.display_type != DestinationDisplayType.folder_only:
            # Render the projected storage space
            if self.display_type == DestinationDisplayType.usage_only:
                y += self.deviceDisplay.dc.padding

            photos_size_to_download, videos_size_to_download = adjusted_download_size(
                photos_size_to_download=self.photos_size_to_download,
                videos_size_to_download=self.videos_size_to_download,
                os_stat_device=self.os_stat_device,
                downloading_to=self._downloading_to,
            )

            details = make_body_details(
                bytes_total=self.storage_space.bytes_total,
                bytes_free=self.storage_space.bytes_free,
                files_to_display=self.files_to_display,
                marked=self.marked,
                photos_size_to_download=photos_size_to_download,
                videos_size_to_download=videos_size_to_download,
            )

            self.deviceDisplay.paint_body(
                painter=painter, x=x, y=y, width=width, details=details
            )

        painter.end()

    @pyqtSlot(int)
    def widthChanged(self, width: int) -> None:
        self.updateGeometry()

    def sizeHint(self) -> QSize:
        if self.display_type == DestinationDisplayType.usage_only:
            height = self.deviceDisplay.dc.padding
        else:
            height = 0

        if self.display_type != DestinationDisplayType.usage_only:
            height += self.deviceDisplay.dc.device_name_height
        if self.display_type != DestinationDisplayType.folder_only:
            height += self.deviceDisplay.dc.storage_height

        return QSize(self.deviceDisplay.width(), height)

    def minimumSize(self) -> QSize:
        return self.sizeHint()

    @pyqtSlot(QMouseEvent)
    def mousePressEvent(self, event: QMouseEvent) -> None:
        if self.menu is None:
            return

        iconRect = self.deviceDisplay.menu_button_rect(0, 0, self.width())

        if iconRect.contains(event.pos()) and event.button() == Qt.LeftButton:
            menuTopReal = iconRect.bottomLeft()
            x = math.ceil(menuTopReal.x())
            y = math.ceil(menuTopReal.y())
            self.setupMenuActions()
            self.menu.popup(self.mapToGlobal(QPoint(x, y)))

    @pyqtSlot(QMouseEvent)
    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """
        Sets the tooltip depending on the position of the mouse.
        """

        if self.menu is None:
            # Relevant only for photo and video destination panels, not the combined
            # storage space display.
            return

        if self.display_type == DestinationDisplayType.folders_and_usage:
            # make tooltip different when hovering above storage space compared
            # to when hovering above the destination folder

            headerRect = QRect(
                0, 0, self.width(), self.deviceDisplay.dc.device_name_height
            )
            if not headerRect.contains(event.pos()):
                if (
                    self.tooltip_display_state
                    != DestinationDisplayTooltipState.storage_space
                ):
                    # Display tooltip for storage space
                    self.setToolTip(self.projected_space_msg)
                    self.tooltip_display_state = (
                        DestinationDisplayTooltipState.storage_space
                    )
                    self.update()
                return

        iconRect = self.deviceDisplay.menu_button_rect(0, 0, self.width())
        if iconRect.contains(event.pos()):
            if self.mouse_pos == DestinationDisplayMousePos.normal:
                self.mouse_pos = DestinationDisplayMousePos.menu

                if self.file_type == FileType.photo:
                    self.setToolTip(_("Configure photo subfolder creation"))
                else:
                    self.setToolTip(_("Configure video subfolder creation"))
                self.tooltip_display_state = DestinationDisplayTooltipState.menu
                self.update()

        else:
            if (
                self.mouse_pos == DestinationDisplayMousePos.menu
                or self.tooltip_display_state != DestinationDisplayTooltipState.path
            ):
                self.mouse_pos = DestinationDisplayMousePos.normal
                self.setToolTip(self.tool_tip)
                self.tooltip_display_state = DestinationDisplayTooltipState.path
                self.update()



#!/usr/bin/env python3

# Copyright (C) 2015-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.


__author__ = "Damon Lynch"
__copyright__ = "Copyright 2015-2024, Damon Lynch"

import contextlib
import locale
import logging
import pickle
import sys

with contextlib.suppress(locale.Error):
    # Use the default locale as defined by the LANG variable
    locale.setlocale(locale.LC_ALL, "")

from PyQt5.QtGui import QGuiApplication

from raphodo.interprocess import (
    DaemonProcess,
    OffloadData,  # noqa: F401
    OffloadResults,
)
from raphodo.proximity import TemporalProximityGroups


class OffloadWorker(DaemonProcess):
    def __init__(self) -> None:
        super().__init__("Offload")

    def run(self) -> None:
        try:
            while True:
                directive, content = self.receiver.recv_multipart()

                self.check_for_command(directive, content)

                data = pickle.loads(content)  # type: OffloadData
                if data.thumbnail_rows:
                    groups = TemporalProximityGroups(
                        thumbnail_rows=data.thumbnail_rows,
                        temporal_span=data.proximity_seconds,
                    )
                    self.content = pickle.dumps(
                        OffloadResults(proximity_groups=groups), pickle.HIGHEST_PROTOCOL
                    )
                    self.send_message_to_sink()
                else:
                    assert data.folders_preview
                    assert data.rpd_files
                    data.folders_preview.generate_subfolders(
                        rpd_files=data.rpd_files, strip_characters=data.strip_characters
                    )
                    self.content = pickle.dumps(
                        OffloadResults(folders_preview=data.folders_preview),
                        pickle.HIGHEST_PROTOCOL,
                    )
                    self.send_message_to_sink()

        except Exception:
            logging.error(
                "An unhandled exception occurred while processing offloaded tasks"
            )
            logging.exception("Traceback:")
        except SystemExit as e:
            sys.exit(e.code)


if __name__ == "__main__":
    # Must initialize QGuiApplication to use QFont() and QFontMetrics
    app = QGuiApplication(sys.argv)

    offload = OffloadWorker()
    offload.run()



#!/usr/bin/env python3

# Copyright (C) 2007-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2007-2024, Damon Lynch"

# ruff: noqa: E402

import datetime
import logging
from typing import Any

import gi

gi.require_version("GExiv2", "0.10")
from gi.repository import GExiv2
from PyQt5.QtCore import QSize

import raphodo.metadata.exiftool as exiftool
import raphodo.metadata.metadataexiftool as metadataexiftool
from raphodo.constants import FileType
from raphodo.utilities import flexible_date_time_parser, image_large_enough_fdo

VENDOR_SERIAL_CODES = (
    "Exif.Photo.BodySerialNumber",
    "Exif.Canon.SerialNumber",
    "Exif.Nikon3.SerialNumber",
    "Exif.OlympusEq.SerialNumber",
    "Exif.Olympus.SerialNumber",
    "Exif.Olympus.SerialNumber2",
    "Exif.Panasonic.SerialNumber",
    "Exif.Fujifilm.SerialNumber",
    "Exif.Image.CameraSerialNumber",
)

VENDOR_SHUTTER_COUNT = (
    "Exif.Nikon3.ShutterCount",
    "Exif.Canon.FileNumber",
    "Exif.Canon.ImageNumber",
)


def photo_date_time(
    metadata: GExiv2.Metadata,
    full_file_name: str | None = None,
    file_type: FileType | None = None,
) -> datetime.datetime | Any:
    """
    Returns in python datetime format the date and time the image was
    recorded.

    Tries these tags, in order:
    Exif.Photo.DateTimeOriginal
    Exif.Image.DateTimeOriginal
    Exif.Image.DateTime

    :return: metadata value, or None if value is not present.
    """

    # GExiv2.Metadata used to provide get_date_time(), but as of version
    # 0.10.09 it appears to have been removed!

    # In any case, get_date_time() seems to have tried only one key,
    # Exif.Photo.DateTimeOriginal
    # Try other keys too, and with a more flexible datetime parser.
    # For example some or maybe all Android 6.0 DNG files use
    # Exif.Image.DateTimeOriginal

    do_log = full_file_name is not None and file_type is not None

    for tag in (
        "Exif.Photo.DateTimeOriginal",
        "Exif.Image.DateTimeOriginal",
        "Exif.Image.DateTime",
    ):
        try:
            dt_string = metadata.get_tag_string(tag)
        except Exception:
            pass
        else:
            if dt_string is None:
                continue

            # ignore all zero values, e.g. '0000:00:00 00:00:00'
            try:
                digits = int("".join(c for c in dt_string if c.isdigit()))
            except ValueError:
                if do_log:
                    logging.warning(
                        "Unexpected malformed date time metadata value %s for photo %s",
                        dt_string,
                        full_file_name,
                    )
            else:
                if not digits:
                    if do_log:
                        logging.debug(
                            "Ignoring date time metadata value %s for photo %s",
                            dt_string,
                            full_file_name,
                        )
                else:
                    try:
                        return datetime.datetime.strptime(
                            dt_string, "%Y:%m:%d %H:%M:%S"
                        )
                    except (ValueError, OverflowError):
                        if do_log:
                            logging.debug(
                                "Error parsing date time metadata %s for photo %s; "
                                "attempting flexible approach",
                                dt_string,
                                full_file_name,
                            )
                        try:
                            dtr, fs = flexible_date_time_parser(dt_string.strip())
                            if do_log:
                                logging.debug(
                                    "Extracted photo time %s using flexible approach",
                                    dtr.strftime(fs),
                                )
                            return dtr
                        except AssertionError:
                            if do_log:
                                logging.warning(
                                    "Error extracting date time metadata '%s' for "
                                    "%s %s",
                                    dt_string,
                                    file_type,
                                    full_file_name,
                                )
                        except (ValueError, OverflowError):
                            if do_log:
                                logging.warning(
                                    "Error parsing date time metadata '%s' for %s %s",
                                    dt_string,
                                    file_type,
                                    full_file_name,
                                )
                        except Exception:
                            if do_log:
                                logging.error(
                                    "Unknown error parsing date time metadata '%s' "
                                    "for %s %s",
                                    dt_string,
                                    file_type,
                                    full_file_name,
                                )
    return None


class MetaData(metadataexiftool.MetadataExiftool, GExiv2.Metadata):
    """
    Provide abstracted access to photo metadata
    """

    def __init__(
        self,
        et_process: exiftool.ExifTool,
        full_file_name: str | None = None,
        raw_bytes: bytearray | None = None,
        app1_segment: bytearray | None = None,
    ) -> None:
        """
        Use GExiv2 to read the photograph's metadata.

        :param et_process: deamon exiftool process
        :param full_file_name: full path of file from which file to read
         the metadata.
        :param raw_bytes: portion of a non-jpeg file from which the
         metadata can be extracted
        :param app1_segment: the app1 segment of a jpeg file, from which
         the metadata can be read
        """

        super().__init__(full_file_name, et_process, FileType.photo)

        self.et_process = et_process

        if full_file_name:
            self.open_path(full_file_name)
        else:
            if raw_bytes is not None:
                self.open_buf(raw_bytes)
            else:
                assert app1_segment is not None
                self.from_app1_segment(app1_segment)

    def _get_rational_components(self, tag: str) -> tuple[Any, Any] | None:
        try:
            x = self.get_exif_tag_rational(tag)
        except Exception:
            return None, None

        try:
            return x.numerator, x.denominator
        except AttributeError:
            try:
                return x.nom, x.den
            except Exception:
                return None, None

    def _get_rational(self, tag: str) -> float | None:
        x, y = self._get_rational_components(tag)
        if x is not None and y is not None:
            return float(x) / float(y)

    def aperture(self, missing="") -> str | Any:
        """
        Returns in string format the floating point value of the image's
        aperture.

        Returns missing if the metadata value is not present.
        """

        a = self._get_rational("Exif.Photo.FNumber")

        return missing if a is None else f"{a:.1f}"

    def iso(self, missing="") -> str | Any:
        """
        Returns in string format the integer value of the image's ISO.

        Returns missing if the metadata value is not present.
        """

        try:
            v = self.get_tag_interpreted_string("Exif.Photo.ISOSpeedRatings")
            if v:
                return v
            else:
                return missing
        except (KeyError, AttributeError):
            return missing

    def _exposure_time_rational(self) -> tuple[Any, Any]:
        return self._get_rational_components("Exif.Photo.ExposureTime")

    def focal_length(self, missing=""):
        """
        Returns in string format the focal length of the lens used to record
        the image.

        Returns missing if the metadata value is not present.
        """

        f = self._get_rational("Exif.Photo.FocalLength")
        if f is not None:
            return "%.0f" % f
        else:
            return missing

    def camera_make(self, missing=""):
        """
        Returns in string format the camera make (manufacturer) used to
        record the image.

        Returns missing if the metadata value is not present.
        """

        try:
            return self.get_tag_string("Exif.Image.Make").strip()
        except (KeyError, AttributeError):
            return missing

    def camera_model(self, missing=""):
        """
        Returns in string format the camera model used to record the image.

        Returns missing if the metadata value is not present.
        """

        try:
            return self.get_tag_string("Exif.Image.Model").strip()
        except (KeyError, AttributeError):
            return missing

    def _fetch_vendor(self, vendor_codes, missing=""):
        for key in vendor_codes:
            try:
                return self.get_tag_string(key).strip()
            except (KeyError, AttributeError):
                pass
        return missing

    def camera_serial(self, missing=""):
        return self._fetch_vendor(VENDOR_SERIAL_CODES, missing)

    def shutter_count(self, missing=""):
        shutter = self._fetch_vendor(VENDOR_SHUTTER_COUNT, missing)
        if shutter != missing:
            return shutter

        if self.full_file_name is None:
            return missing

        if self.camera_make().lower() == "sony":
            try:
                ic = self.et_process.get_tags(["ImageCount"], self.full_file_name)
            except (ValueError, TypeError):
                return missing
            if ic:
                return ic.get("ImageCount", missing)

        return missing

    def file_number(self, missing=""):
        """
        Returns Exif.CanonFi.FileNumber, not to be confused with
        Exif.Canon.FileNumber.

        Uses ExifTool to extract the value, because the exiv2
        implementation is currently problematic

        See:
        https://bugs.launchpad.net/rapid/+bug/754531
        """
        if "Exif.CanonFi.FileNumber" in self and self.full_file_name is not None:
            assert self.et_process is not None
            return super().file_number(missing)

    def owner_name(self, missing=""):
        try:
            return self.get_tag_string("Exif.Canon.OwnerName").strip()
        except (KeyError, AttributeError):
            return missing

    def copyright(self, missing=""):
        try:
            return self.get_tag_string("Exif.Image.Copyright").strip()
        except (KeyError, AttributeError):
            return missing

    def artist(self, missing=""):
        try:
            return self.get_tag_string("Exif.Image.Artist").strip()
        except (KeyError, AttributeError):
            return missing

    def date_time(
        self,
        missing: str | None = "",
        ignore_file_modify_date: bool | None = False,
    ) -> datetime.datetime | Any:
        """
        Returns in python datetime format the date and time the image was
        recorded.

        Tries these tags, in order:
        Exif.Photo.DateTimeOriginal
        Exif.Image.DateTimeOriginal
        Exif.Image.DateTime

        :return: metadata value, or missing if value is not present.
        """

        dt = photo_date_time(
            metadata=self, full_file_name=self.full_file_name, file_type=self.file_type
        )

        if dt is None:
            return missing
        else:
            return dt

    def sub_seconds(self, missing="00") -> str | Any:
        """
        Returns the subsecond the image was taken, as recorded by the
        camera
        """

        try:
            return self.get_tag_string("Exif.Photo.SubSecTimeOriginal")
        except (KeyError, AttributeError):
            return missing

    def orientation(self, missing="") -> str | Any:
        """
        Returns the orientation of the image, as recorded by the camera
        Return type int
        """

        try:
            return self.get_tag_string("Exif.Image.Orientation")
        except (KeyError, AttributeError):
            return missing

    def get_small_thumbnail(self) -> bytes:
        """
        Get the small thumbnail image (if it exists)
        :return: thumbnail image in raw bytes (could be zero bytes)
        """

        return self.get_exif_thumbnail()

    def get_indexed_preview(self) -> bytes | None:
        """
        Extract preview image from the metadata

        :param preview_number: which preview to get
        :return: preview image in raw bytes, if found, else None
        """

        previews = self.get_preview_properties()
        if previews:
            # In every RAW file I've analyzed, the smallest preview is always first
            for preview in previews:
                data = self.get_preview_image(preview).get_data()
                if data:
                    return data
        logging.warning("Photo %s has no image previews", self.full_file_name)
        return None

    def get_small_thumbnail_or_first_indexed_preview(self) -> bytes | None:
        """
        First attempt to get the small thumbnail image. If it does not exist,
        extract the smallest preview image from the metadata

        :return: thumbnail / preview image in raw bytes, if found, else None
        """

        # Look for Thumbnail Image if the file format supports it
        if self.ext not in self.preview_smallest or self.ext in self.may_have_thumbnail:
            thumbnail = self.get_small_thumbnail()
            if thumbnail:
                return thumbnail

        # Otherwise look for the smallest preview image for this format
        return self.get_indexed_preview()

    def get_preview_256(self) -> bytes | None:
        """
        :return: if possible, return a preview image that is preferrably larger than
         256 pixels, else the smallest preview if it exists
        """

        previews = self.get_preview_properties()
        if not previews:
            return None

        for preview in previews:
            if image_large_enough_fdo(
                QSize(preview.get_width(), preview.get_height())
            ) and not (
                self.ext in self.ignore_tiff_preview_256
                and preview.get_mime_type() == "image/tiff"
            ):
                break

        # At this point we have a preview that may or may not be bigger than 160x120.
        # On older RAW files, no. On newer RAW files, yes.
        return self.get_preview_image(preview).get_data()


class DummyMetaData(MetaData):
    """
    Class which gives metadata values for an imaginary photo.

    Useful for displaying in preference examples etc. when no image is ready to
    be downloaded.

    See MetaData class for documentation of class methods.
    """

    def __init__(self):
        pass

    def readMetadata(self):
        pass

    def aperture(self, missing=""):
        return "2.0"

    def iso(self, missing=""):
        return "100"

    def exposure_time(self, alternative_format=False, missing=""):
        if alternative_format:
            return "4000"
        else:
            return "1/4000"

    def focal_length(self, missing=""):
        return "135"

    def camera_make(self, missing=""):
        return "Canon"

    def camera_model(self, missing=""):
        return "Canon EOS 5D"

    def short_camera_model(self, include_characters="", missing=""):
        return "5D"

    def camera_serial(self, missing=""):
        return "730402168"

    def shutter_count(self, missing=""):
        return "387"

    def owner_name(self, missing=""):
        return "Photographer Name"

    def date_time(self, missing="", ignore_file_modify_date=False):
        return datetime.datetime.now()

    def subSeconds(self, missing="00"):
        return "57"

    def orientation(self, missing=""):
        return 1

    def file_number(self, missing=""):
        return "428"


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print("Usage: " + sys.argv[0] + " path/to/photo/containing/metadata")
        m = DummyMetaData()
        et_process = None
    else:
        et_process = exiftool.ExifTool()
        et_process.start()
        m = MetaData(full_file_name=sys.argv[1], et_process=et_process)

    print("f" + m.aperture("missing "))
    print("ISO " + m.iso("missing "))
    print(m.exposure_time(missing="missing ") + " sec")
    print(m.exposure_time(alternative_format=True, missing="missing "))
    print(m.focal_length("missing ") + "mm")
    print(m.camera_make())
    print(m.camera_model())
    print(m.short_camera_model())
    print(m.short_camera_model(include_characters="\-"))
    print(m.date_time())
    print(m.orientation())
    print("Serial number:", m.camera_serial(missing="missing"))
    print("Shutter count:", m.shutter_count())
    print("Subseconds:", m.sub_seconds(), type(m.sub_seconds()))
    print("File number:", m.file_number())
    preview = m.get_small_thumbnail_or_first_indexed_preview()
    if m is not None:
        print("Preview size", len(preview))
    else:
        print("Preview not availabe")

    if et_process is not None:
        et_process.terminate()



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Display file system folders and allow the user to select one
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"

import logging
import os
import pathlib
import re

from PyQt5.QtCore import (
    QDir,
    QItemSelectionModel,
    QModelIndex,
    QPoint,
    QSize,
    QSortFilterProxyModel,
    Qt,
    pyqtSignal,
    pyqtSlot,
)
from PyQt5.QtGui import QFont, QPainter
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QAction,
    QFileSystemModel,
    QMenu,
    QSizePolicy,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QTreeView,
)
from showinfm import show_in_file_manager

from raphodo.constants import (
    Roles,
    filtered_file_browser_directories,
    minFileSystemViewHeight,
    minPanelWidth,
    non_system_root_folders,
)
from raphodo.storage.storage import get_media_dir, gvfs_gphoto2_path
from raphodo.ui.viewutils import (
    TopFramedVerticalScrollBar,
    darkModeIcon,
    standard_font_size,
)
from raphodo.wslutils import wsl_filter_directories


class FileSystemModel(QFileSystemModel):
    """
    Use Qt's built-in functionality to model the file system.

    Augment it by displaying provisional subfolders in the photo and video
    download destinations.
    """

    def __init__(self, parent) -> None:
        super().__init__(parent)

        # More filtering done in the FileSystemFilter
        self.setFilter(QDir.AllDirs | QDir.NoDotAndDotDot)

        s = standard_font_size()
        size = QSize(s, s)

        self.folder_icon = darkModeIcon(
            path=":/icons/folder.svg", size=size, soften_regular_mode_color=True
        )
        self.download_folder_icon = darkModeIcon(
            path=":/icons/folder-filled.svg", size=size, soften_regular_mode_color=True
        )

        self.setRootPath("/")

        # The next two values are set via FolderPreviewManager.update()
        # They concern provisional folders that will be used if the
        # download proceeds, and all files are downloaded.

        # First value: subfolders we've created to demonstrate to the user
        # where their files will be downloaded to
        self.preview_subfolders = set()  # type: set[str]
        # Second value: subfolders that already existed, but that we still
        # want to indicate to the user where their files will be downloaded to
        self.download_subfolders = set()  # type: set[str]

        # Folders that were actually used to download files into
        self.subfolders_downloaded_into = set()  # type: set[str]

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        if role == Qt.DecorationRole:
            path = index.data(QFileSystemModel.FilePathRole)  # type: str
            if (
                path in self.download_subfolders
                or path in self.subfolders_downloaded_into
            ):
                return self.download_folder_icon
            else:
                return self.folder_icon
        if role == Roles.folder_preview:
            path = index.data(QFileSystemModel.FilePathRole)
            return (
                path in self.preview_subfolders
                and path not in self.subfolders_downloaded_into
            )

        return super().data(index, role)

    def add_subfolder_downloaded_into(self, path: str, download_folder: str) -> bool:
        """
        Add a path to the set of subfolders that indicate where files where
        downloaded.
        :param path: the full path to the folder
        :return: True if the path was not added before, else False
        """

        if path not in self.subfolders_downloaded_into:
            self.subfolders_downloaded_into.add(path)

            pl_subfolders = pathlib.Path(path)
            pl_download_folder = pathlib.Path(download_folder)

            for subfolder in pl_subfolders.parents:
                if pl_download_folder not in subfolder.parents:
                    break
                self.subfolders_downloaded_into.add(str(subfolder))
            return True
        return False


class FileSystemView(QTreeView):

    showSystemFolders = pyqtSignal(bool)
    filePathReset = pyqtSignal()

    def __init__(self, model: FileSystemModel, rapidApp, parent=None) -> None:
        super().__init__(parent)
        self.rapidApp = rapidApp
        self.fileSystemModel = model
        self.setHeaderHidden(True)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        self.setMinimumWidth(minPanelWidth())
        self.setMinimumHeight(minFileSystemViewHeight())
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onCustomContextMenu)
        self.contextMenu = QMenu()
        self.openInFileBrowserAct = self.contextMenu.addAction(
            _("Open in File Browser...")
        )
        self.openInFileBrowserAct.triggered.connect(self.doOpenInFileBrowserAct)
        self.openInFileBrowserAct.setEnabled(self.rapidApp.file_manager is not None)
        self.clickedIndex = None  # type: QModelIndex | None

        self.resetSelectionAct = self.contextMenu.addAction(_("Reset"))
        self.resetSelectionAct.triggered.connect(self.doResetSelectionAct)

        self.showSystemFoldersAct = QAction(
            _("Show System Folders"),
            self,
            enabled=True,
            checkable=True,
            triggered=self.doShowSystemFoldersAct,
        )
        self.contextMenu.addAction(self.showSystemFoldersAct)

        self.setVerticalScrollBar(TopFramedVerticalScrollBar(name="fileSystemView"))

    def hideColumns(self) -> None:
        """
        Call only after the model has been initialized
        """
        for i in (1, 2, 3):
            self.hideColumn(i)

    def goToPath(self, path: str, scrollTo: bool = True) -> None:
        """
        Select the path, expand its subfolders, and scroll to it
        :param path:
        :return:
        """
        if not path:
            return
        index = self.model().mapFromSource(self.fileSystemModel.index(path))
        self.setExpanded(index, True)
        selection = self.selectionModel()
        selection.select(
            index, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
        )
        if scrollTo:
            self.scrollTo(index, QAbstractItemView.PositionAtTop)

    def expandPreviewFolders(self, path: str) -> bool:
        """
        Expand any unexpanded preview folders

        :param path: path under which to expand folders
        :return: True if path was expanded, else False
        """

        self.goToPath(path, scrollTo=True)
        if not path:
            return False

        expanded = False
        for path in self.fileSystemModel.download_subfolders:
            # print('path', path)
            index = self.model().mapFromSource(self.fileSystemModel.index(path))
            if not self.isExpanded(index):
                self.expand(index)
                expanded = True
        return expanded

    def expandPath(self, path) -> None:
        index = self.model().mapFromSource(self.fileSystemModel.index(path))
        if not self.isExpanded(index):
            self.expand(index)

    def onCustomContextMenu(self, point: QPoint) -> None:
        index = self.indexAt(point)
        self.showSystemFoldersAct.setChecked(self.rapidApp.prefs.show_system_folders)
        if index.isValid():
            self.clickedIndex = index
            self.openInFileBrowserAct.setEnabled(True)
        else:
            self.openInFileBrowserAct.setEnabled(False)
        self.showSystemFoldersAct.setEnabled(
            not self.rapidApp.prefs.source_or_destination_is_system_folder()
        )
        self.contextMenu.exec(self.mapToGlobal(point))

    @pyqtSlot()
    def doOpenInFileBrowserAct(self) -> None:
        index = self.clickedIndex
        if index:
            uri = self.fileSystemModel.filePath(index.model().mapToSource(index))
            logging.debug(
                "Calling show_in_file_manager() with %s and %s",
                self.rapidApp.file_manager,
                uri,
            )
            show_in_file_manager(path_or_uri=uri, open_not_select_directory=True)

    @pyqtSlot()
    def doShowSystemFoldersAct(self) -> None:
        self.showSystemFolders.emit(self.showSystemFoldersAct.isChecked())

    @pyqtSlot()
    def doResetSelectionAct(self) -> None:
        self.selectionModel().clear()
        self.filePathReset.emit()


class FileSystemFilter(QSortFilterProxyModel):
    """
    Filter out the display of RPD's cache and temporary directories, in addition to
    a set of standard directories that should not be displayed.
    """

    filterInvalidated = pyqtSignal()

    def __init__(self, parent: "RapidWindow" = None):  # noqa: F821
        super().__init__(parent)
        self.is_wsl2 = parent.is_wsl2
        self.prefs = parent.prefs
        if self.is_wsl2:
            self.filter_paths = wsl_filter_directories()
            # Filter out system created WSL working directories
            self.regex = re.compile(r"/wsl[\w]")
        else:
            self.filter_paths = set()
        self.filtered_dir_names = filtered_file_browser_directories
        self.non_system_root_folders = non_system_root_folders
        if get_media_dir().startswith("/run"):
            self.non_system_root_folders.append("/run")

    def setTempDirs(self, dirs: list[str]) -> None:
        filters = [os.path.basename(path) for path in dirs]
        self.filtered_dir_names = self.filtered_dir_names | set(filters)
        self.invalidateFilter()

    def filterAcceptsRow(
        self, sourceRow: int, sourceParent: QModelIndex = None
    ) -> bool:
        index = self.sourceModel().index(
            sourceRow, 0, sourceParent
        )  # type: QModelIndex
        path = index.data(QFileSystemModel.FilePathRole)  # type: str

        if not self.prefs.show_system_folders and path != "/":
            path_ok = False
            for folder in self.non_system_root_folders:
                if path.startswith(folder):
                    path_ok = True
                    break
            if not path_ok:
                return False

        if gvfs_gphoto2_path(path):
            logging.debug("Rejecting browsing path %s", path)
            return False

        if not self.filtered_dir_names and not self.is_wsl2:
            return True

        file_name = index.data(QFileSystemModel.FileNameRole)
        do_filter = (
            file_name not in self.filtered_dir_names and path not in self.filter_paths
        )

        if self.is_wsl2:
            do_filter = do_filter and self.regex.match(path) is None
        return do_filter

    @pyqtSlot(bool)
    def setShowSystemFolders(self, enabled: bool) -> None:
        self.prefs.show_system_folders = enabled
        self.invalidateFilter()
        self.filterInvalidated.emit()


class FileSystemDelegate(QStyledItemDelegate):
    """
    Italicize provisional download folders that were not already created
    """

    def __init__(self, parent=None):
        super().__init__(parent)

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        if index is None:
            return

        folder_preview = index.data(Roles.folder_preview)
        if folder_preview:
            font = QFont()
            font.setItalic(True)
            option.font = font

        super().paint(painter, option, index)
